import abc
from dataclasses import dataclass
from typing import Tuple, Union, Callable, List

from tensorflow import keras


@dataclass
class ModelBase(abc.ABC):
    def __init__(self, observation_shape: List[int], n_actions: int, output_activation: Union[None, str] = None,
                 unit_scale: int = 1, learning_rate: float = 0.0001, opt: str = 'Adam') -> None:
        """
        :param observation_shape: Tuple specifying input shape.
        :param n_actions: Int specifying number of outputs
        :param output_activation: Activation function for output. Eg. None for value estimation (off-policy methods) or
                                  'softmax' for action probabilities (on-policy methods).
        :param unit_scale: Multiplier for all units in FC layers in network. Default 1 = 16 units for first layer,
                            8 for second.
        :param opt: Keras optimiser to use. Should be string. This is to avoid storing TF/Keras objects here.
        :param learning_rate: Learning rate for optimiser.

        """
        self.observation_shape = observation_shape
        self.n_actions = n_actions
        self.unit_scale = unit_scale
        self.output_activation = output_activation
        self.learning_rate = learning_rate
        self.opt = opt

    def compile(self, model_name: str = 'model', loss: Union[str, Callable] = 'mse') -> keras.Model:
        """
        Compile a copy of the model using the provided loss.

        Note loss is added here to avoid storing in self. We don't want to do that, as if this model is pickled deepcopy
        will be disabled for TF objects if eager mode is disabled. It's better to use as needed rather than storing.

        :param model_name: Name of model
        :param loss: Model loss. Default 'mse'. Can be custom callable.
        """
        # Get optimiser
        if self.opt.lower() == 'adam':
            opt = keras.optimizers.Adam
        elif self.opt.lower() == 'rmsprop':
            opt = keras.optimizers.RMSprop
        else:
            raise ValueError(f"Invalid optimiser {self.opt}")

        state_input, action_output = self._model_architecture()
        model = keras.Model(inputs=[state_input], outputs=[action_output], name=model_name)
        model.compile(optimizer=opt(learning_rate=self.learning_rate), loss=loss)

        return model

    @abc.abstractmethod
    def _model_architecture(self) -> Tuple[keras.layers.Layer, keras.layers.Layer]:
        """Define model construction function. Should return input layer and output layer."""
        pass

    def plot(self, model_name: str = 'model') -> None:
        keras.utils.plot_model(self.compile(model_name), to_file=f"{model_name}.png", show_shapes=True)



import copy
from typing import List

import numpy as np
from numpy.testing import assert_array_almost_equal

from reinforcement_learning_keras.agents.q_learning.linear_q_agent import LinearQAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from tests.unit.agents.random.test_random_agent import TestRandomAgent


class TestLinearQAgent(TestRandomAgent):
    _sut = LinearQAgent
    _config = CartPoleConfig(agent_type='linear_q', plot_during_training=False)

    def _agent_specific_set_up(self):
        """This agent requires no specific set up, however this method should still be set."""
        pass

    @staticmethod
    def _checkpoint_model(agent: LinearQAgent) -> List[np.ndarray]:
        """Get coefs from each model"""
        return [copy.deepcopy(m.coef_) for m in agent.mods.values()]

    def _assert_model_unchanged(self, agent: LinearQAgent, checkpoint: List[np.ndarray]) -> None:
        for previous_coefs, current_coefs in zip([m.coef_ for m in agent.mods.values()], checkpoint):
            assert_array_almost_equal(previous_coefs, current_coefs)

    def _assert_model_changed(self, agent: LinearQAgent, checkpoint: List[np.ndarray]) -> None:
        changes_to_single_action_model = []
        for previous_coefs, current_coefs in zip([m.coef_ for m in agent.mods.values()], checkpoint):
            changes_to_single_action_model.append(np.any(np.not_equal(previous_coefs.round(8),
                                                                      current_coefs.round(8))))

        self.assertTrue(np.any(changes_to_single_action_model))

    def _assert_relevant_play_episode_change(self, agent: LinearQAgent, checkpoint: List[np.ndarray]) -> None:
        """This can differ between MC and TD agents. In MC case model might not be updated but buffer is."""
        self._assert_model_changed(agent, checkpoint)

    def _assert_relevant_after_play_episode_change(self, agent: LinearQAgent, checkpoint: List[np.ndarray]) -> None:
        self._assert_model_changed(agent, checkpoint)

    def _assert_agent_ready(self, agent: LinearQAgent) -> None:
        self.assertIsNotNone(agent.env_builder._env)

    def test_model_set_during_init(self) -> None:
        # Act
        agent = self._ready_agent()

        # Assert
        self.assertIsInstance(agent.mods, dict)
        self.assertIsNotNone(agent.mods[0])
        self.assertIsNotNone(agent.mods[1])


del TestRandomAgent



import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

from reinforcement_learning_keras.agents.components.replay_buffers.continuous_buffer import ContinuousBuffer


class TestContinuousBuffer(unittest.TestCase):
    _sut = ContinuousBuffer

    def _fill_replay_buffer(self, shape=(5,)) -> ContinuousBuffer:
        replay_buffer = self._sut()
        for i in range(10):
            replay_buffer.append((np.zeros(shape=shape) + i,
                                  i, 0.9, False))

        return replay_buffer

    def _fill_replay_buffer_include_dones(self, shape=(5,)) -> ContinuousBuffer:
        replay_buffer = self._sut()
        dones = np.array([False] * 10)
        dones[6] = True
        for i, d in zip(range(10), dones):
            replay_buffer.append((np.zeros(shape=shape) + i,
                                  i, 0.9, d))

        return replay_buffer

    def test_get_sample_with_1_dim_state_returns_expected_shapes(self) -> None:
        # Arrange
        shape = (5,)
        rb = self._fill_replay_buffer()

        # Act
        ss, aa, rr, dd, ss_ = rb.sample_batch(2)

        # Assert
        ss_0 = [np.unique(ss[0][0])]
        ss__0 = [np.unique(ss_[0][0])]
        self.assertListEqual([s + 1 for s in ss_0], ss__0)

        self.assertIsInstance(aa, list)
        self.assertIsInstance(rr, list)
        self.assertEqual(shape, ss[0].shape)

    def test_sequential_get_batch_returns_sequential_samples(self) -> None:
        # Arrange
        rb = self._fill_replay_buffer(shape=(5, 5, 3))

        # Act
        ss, aa, rr, dd, ss_ = rb.get_batch(idxs=[4, 5])

        # Assert
        assert_array_almost_equal(ss[1], ss_[0])

    def test_get_sample_with_multi_dim_state_returns_expected_shapes(self) -> None:
        # Arrange
        shape = (5, 5, 3)
        rb = self._fill_replay_buffer(shape=shape)

        # Act
        ss, aa, rr, dd, ss_ = rb.sample_batch(8)

        # Assert
        ss_0 = [np.unique(ss[0][0])]
        ss__0 = [np.unique(ss_[0][0])]
        self.assertListEqual([s + 1 for s in ss_0], ss__0)
        self.assertEqual(shape, ss[0].shape)



import copy
from typing import List
from unittest.mock import patch

import numpy as np
from numpy.testing import assert_array_almost_equal

from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from tests.unit.agents.random.test_random_agent import TestRandomAgent


class TestDeepQAgent(TestRandomAgent):
    _sut = DeepQAgent
    _config = CartPoleConfig(agent_type='dqn', plot_during_training=False)

    def _agent_specific_set_up(self):
        VirtualGPU(256)

        # (in .play_episode)
        self._expected_model_update_after_training_episode: int = 0
        self._expected_model_update_after_playing_episode: int = 0

    def _ready_agent(self) -> DeepQAgent:
        agent = self._sut(**self._config.build())
        # Sync weights so checks are simpler
        agent.update_target_model()

        return agent

    @staticmethod
    def _checkpoint_model(agent: DeepQAgent) -> List[np.ndarray]:
        """Get coefs from each model"""
        return copy.deepcopy(agent._action_model.get_weights())

    def _assert_model_unchanged(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        action_weights = agent._action_model.get_weights()
        value_weights = agent._target_model.get_weights()
        for w in range(len(action_weights)):
            assert_array_almost_equal(action_weights[w], checkpoint[w])
            assert_array_almost_equal(value_weights[w], checkpoint[w])

    def _assert_relevant_play_episode_change(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        self._assert_buffer_changed(agent, checkpoint)

    def _assert_relevant_after_play_episode_change(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        """TODO: self._assert_value_model_changed(agent, checkpoint) (?)"""
        pass

    def _assert_buffer_changed(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        # TODO
        pass

    def _assert_action_model_changed(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        action_weights = agent._action_model.get_weights()
        for w in range(len(checkpoint)):
            self.assertFalse(np.all(action_weights[w].round(6) == checkpoint[w].round(6)))

    def _assert_value_model_changed(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        value_weights = agent._target_model.get_weights()
        for w in range(len(checkpoint)):
            self.assertFalse(np.all(value_weights[w].round(6) == checkpoint[w].round(6)))

    def _assert_model_changed(self, agent: DeepQAgent, checkpoint: List[np.ndarray]) -> None:
        self._assert_action_model_changed(agent, checkpoint)
        self._assert_value_model_changed(agent, checkpoint)

    def _assert_agent_unready(self, agent: DeepQAgent) -> None:
        self.assertIsNone(agent._action_model)
        self.assertIsNone(agent._target_model)
        self.assertIsNone(agent.replay_buffer)
        self.assertFalse(agent.ready)

    def _assert_agent_ready(self, agent: DeepQAgent) -> None:
        self.assertIsNotNone(agent._action_model)
        self.assertIsNotNone(agent._target_model)
        self.assertIsNotNone(agent.replay_buffer)
        self.assertTrue(agent.ready)

    def test_play_episode_steps_does_not_call_update_models_when_not_training(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        with patch.object(agent, 'update_model') as mocked_update_model, \
                patch.object(agent, "update_target_model") as mocked_update_value_model:
            _ = agent.play_episode(max_episode_steps=self._n_step, training=False, render=False)

        # Assert
        self.assertEqual(self._expected_model_update_during_playing_episode, mocked_update_model.call_count)
        self.assertEqual(self._expected_model_update_after_playing_episode, mocked_update_value_model.call_count)
        self.assertEqual(self._n_step, agent.env_builder._env._max_episode_steps)

    def test_play_episode_steps_calls_update_models_when_training(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        with patch.object(agent, 'update_model') as mocked_update_model, \
                patch.object(agent, "update_target_model") as mocked_update_value_model:
            _ = agent.play_episode(max_episode_steps=self._n_step, training=True, render=False)

        # Assert
        self.assertEqual(self._expected_model_update_during_training_episode, mocked_update_model.call_count)
        self.assertEqual(self._expected_model_update_after_training_episode, mocked_update_value_model.call_count)
        self.assertEqual(self._n_step, agent.env_builder._env._max_episode_steps)


del TestRandomAgent



from typing import Any, Callable

from tqdm import tqdm


class TQDMHandler:
    """
    Turns TQDM waitbars on or off depending on a verbosity setting.

    Useful to avoid having to change the for loop manually to switch.
    """

    def __init__(self):
        self.tqdm_runner: Callable = tqdm

    @staticmethod
    def _fake_tqdm(x: Any) -> Any:
        return x

    def set_tqdm(self, verbose: bool = False) -> None:
        """Turn tqdm on or of depending on verbosity setting."""
        if verbose:
            self.tqdm_runner = tqdm
        else:
            self.tqdm_runner = self._fake_tqdm



from typing import Tuple

from tensorflow import keras
from tensorflow.keras import backend as K

from reinforcement_learning_keras.enviroments.model_base import ModelBase


class SmallDuelingNN(ModelBase):

    def _model_architecture(self) -> Tuple[keras.layers.Layer, keras.layers.Layer]:
        n_units = 16 * self.unit_scale

        state_input = keras.layers.Input(name='input', shape=self.observation_shape)
        fc1 = keras.layers.Dense(units=int(n_units), name='fc1', activation='relu')(state_input)
        fc2 = keras.layers.Dense(units=int(n_units / 2), name='fc2', activation='relu')(fc1)

        # Separate layers for baseline value (1 node) and action advantages (n action nodes)
        v_layer = keras.layers.Dense(1, activation='linear')(fc2)
        a_layer = keras.layers.Dense(self.n_actions, activation='linear')(fc2)

        # Function definition for lambda layer. Base value layer + actions value layer - mean action values
        def _merge_layer(layer_inputs: list):
            return layer_inputs[0] + layer_inputs[1] - K.mean(layer_inputs[1], axis=1, keepdims=True)

        action_output = keras.layers.Lambda(_merge_layer, output_shape=(self.n_actions,),
                                            name="output")([v_layer, a_layer])

        return state_input, action_output



"""Train and few LinearQAgents, plot the results, and run an episode on the best agent."""

from reinforcement_learning_keras.agents.q_learning.linear_q_agent import LinearQAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig

from reinforcement_learning_keras.experiment.agent_experiment import AgentExperiment


def run_exp(n_episodes: int = 1000, max_episode_steps: int = 500):
    exp = AgentExperiment(agent_class=LinearQAgent,
                          agent_config=CartPoleConfig('linear_q'),
                          n_reps=32,
                          n_jobs=32,
                          training_options={"n_episodes": n_episodes,
                                            "max_episode_steps": max_episode_steps})

    exp.run()
    exp.save(fn=f"{LinearQAgent.__name__}_experiment.pkl")


if __name__ == "__main__":
    run_exp()



"""Train and few LinearQAgents, plot the results, and run an episode on the best agent."""
from reinforcement_learning_keras.agents.q_learning.linear_q_agent import LinearQAgent
from reinforcement_learning_keras.enviroments.mountain_car.mountain_car_config import MountainCarConfig
from reinforcement_learning_keras.experiment.agent_experiment import AgentExperiment


def run_exp(n_episodes: int = 500, max_episode_steps: int = 1000):
    exp = AgentExperiment(agent_class=LinearQAgent,
                          agent_config=MountainCarConfig(agent_type='linear_1'),
                          n_reps=6,
                          n_jobs=6,
                          training_options={"n_episodes": n_episodes,
                                            "max_episode_steps": max_episode_steps})

    exp.run()
    exp.save(fn=f"{LinearQAgent.__name__}_experiment.pkl")


if __name__ == "__main__":
    run_exp(n_episodes=500)



from dataclasses import dataclass
from typing import Any, Dict, Tuple, Iterable, Callable

from reinforcement_learning_keras.agents.agent_base import AgentBase
from reinforcement_learning_keras.agents.components.helpers.env_builder import EnvBuilder
from reinforcement_learning_keras.agents.components.history.training_history import TrainingHistory
from reinforcement_learning_keras.agents.random.random_model import RandomModel
from reinforcement_learning_keras.enviroments.config_base import ConfigBase


@dataclass
class RandomAgent(AgentBase):
    """
    A non-abstract agent implementing AgentBase interface but acts randomly and learns nothing.

    Useful as a baseline and for testing.
    """
    env_spec: str
    training_history: TrainingHistory
    env_wrappers: Iterable[Callable] = ()
    name: str = 'RandomAgent'

    def __post_init__(self) -> None:
        self.env_builder = EnvBuilder(self.env_spec, self.env_wrappers)
        self._build_model()

    def __getstate__(self) -> Dict[str, Any]:
        return self._pickle_compatible_getstate()

    def _build_model(self) -> None:
        """Set model function. Note using a lambda breaks pickle support."""
        self.model = RandomModel(self.env.action_space.n)

    def update_model(self, *args, **kwargs) -> None:
        """No model to update."""
        pass

    def get_action(self, s: Any, **kwargs) -> int:
        return self.model.predict()

    def _play_episode(self, max_episode_steps: int = 500,
                      training: bool = False, render: bool = True) -> Tuple[float, int]:
        """
        Play a single episode and return the total reward.

        :param max_episode_steps: Max steps before stopping, overrides any time limit set by Gym.
        :param training: Bool to indicate whether or not to use this experience to update the model.
        :param render: Bool to indicate whether or not to call env.render() each training step.
        :return: The total real reward for the episode.
        """
        self.env._max_episode_steps = max_episode_steps
        _ = self.env.reset()
        total_reward = 0
        for frame in range(max_episode_steps):
            action = self.get_action(None)
            _, reward, done, _ = self.env.step(action)
            total_reward += reward

            if render:
                self.env.render()

            if done:
                break

        return total_reward, frame

    @classmethod
    def example(cls, config: ConfigBase, render: bool = True,
                n_episodes: int = 500, max_episode_steps: int = 500) -> "RandomAgent":
        """Create, train, and save agent for a given config."""
        config_dict = config.build()

        agent = cls(**config_dict)

        agent.train(verbose=True, render=render,
                    n_episodes=n_episodes, max_episode_steps=max_episode_steps)
        agent.save()

        return agent


if __name__ == "__main__":
    from reinforcement_learning_keras.enviroments import PongConfig
    from reinforcement_learning_keras.enviroments.cart_pole import CartPoleConfig
    from reinforcement_learning_keras.enviroments import MountainCarConfig

    agent_mountain_car = RandomAgent.example(
        MountainCarConfig(agent_type='random', plot_during_training=True), max_episode_steps=1500, render=False)

    agent_cart_pole = RandomAgent.example(CartPoleConfig(agent_type='random', plot_during_training=True))
    agent_mountain_car = RandomAgent.example(MountainCarConfig(agent_type='random', plot_during_training=True))
    agent_pong = RandomAgent.example(PongConfig(agent_type='random', plot_during_training=True),
                                     max_episode_steps=10000)



import gym
import numpy as np


class SqueezeWrapper(gym.ObservationWrapper):
    """Remove any extra dimensions (eg. row) that might have been added by wrappers using sklearn components."""
    def observation(self, obs: np.ndarray) -> np.ndarray:
        return obs.squeeze()



import os
from functools import partial
from typing import Any, Dict

from reinforcement_learning_keras.agents.components.replay_buffers.continuous_buffer import ContinuousBuffer
from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy
from reinforcement_learning_keras.enviroments.atari.environment_processing.fire_start_wrapper import FireStartWrapper
from reinforcement_learning_keras.enviroments.atari.environment_processing.frame_buffer_wrapper import \
    FrameBufferWrapper
from reinforcement_learning_keras.enviroments.atari.environment_processing.image_process_wrapper import \
    ImageProcessWrapper
from reinforcement_learning_keras.enviroments.atari.environment_processing.max_and_skip_wrapper import MaxAndSkipWrapper
from reinforcement_learning_keras.agents.models.conv_nn import ConvNN
from reinforcement_learning_keras.agents.models.dueling_conv_nn import DuelingConvNN
from reinforcement_learning_keras.enviroments.config_base import ConfigBase


class AtariDefaultConfig(ConfigBase):
    """Defines configs for Pong."""
    env_spec: str
    supported_agents = ('dqn', 'double_dqn', 'dueling_dqn', 'double_dueling_dqn', 'random')
    supported_modes = ('diff', 'stack')
    gpu_memory: int = 2048

    _wrappers_stack = (MaxAndSkipWrapper, ImageProcessWrapper, FireStartWrapper, FrameBufferWrapper)
    _wrappers_diff = (MaxAndSkipWrapper, ImageProcessWrapper, FireStartWrapper,
                      partial(FrameBufferWrapper, buffer_length=2, buffer_function='diff'))

    def __init__(self, mode: str = 'diff', *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if mode not in self.supported_modes:
            raise ValueError(f"Mode {mode} is not a supported mode ({self.supported_modes})")

        self.mode = mode
        if self.mode == "diff":
            self.env_wrappers = self._wrappers_diff
            self.frame_depth = 1
            self.wrapped_env = FrameBufferWrapper(
                FireStartWrapper(ImageProcessWrapper(MaxAndSkipWrapper(self.unwrapped_env))),
                buffer_length=2, buffer_function='diff')
        if self.mode == "stack":
            self.env_wrappers = self._wrappers_stack
            self.frame_depth = 3
            self.wrapped_env = FrameBufferWrapper(FireStartWrapper(
                ImageProcessWrapper(MaxAndSkipWrapper(self.unwrapped_env))),
                buffer_length=3, buffer_function='stack')

    def _build_for_dqn(self) -> Dict[str, Any]:
        return {'name': os.path.join(self.folder, 'DeepQAgent'),
                'env_spec': self.env_spec,
                'env_wrappers': self.env_wrappers,
                'model_architecture': ConvNN(observation_shape=(84, 84, self.frame_depth), n_actions=6,
                                             output_activation=None, opt='adam', learning_rate=0.000105),
                'gamma': 0.99,
                'final_reward': None,
                # Use eps_initial > 1 here so only random actions used for first steps, which will make filling the
                # replay buffer more efficient. It'll also avoid decaying eps while not training.
                # Alternative: 'eps': EpsilonGreedy(eps_initial=1.2, decay=0.000025, eps_min=0.01,
                #                                   decay_schedule='compound'),
                'eps': EpsilonGreedy(eps_initial=1.1, decay=0.00001, eps_min=0.01, decay_schedule='linear'),
                'replay_buffer': ContinuousBuffer(buffer_size=10000),
                'replay_buffer_samples': 32}

    def _build_for_dueling_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DuelingDQN'),
                            'model_architecture': DuelingConvNN(observation_shape=(84, 84, self.frame_depth),
                                                                n_actions=6, opt='adam', learning_rate=0.000102)})

        return config_dict

    def _build_for_double_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DoubleDQN'),
                            'double': True,
                            'model_architecture': ConvNN(observation_shape=(84, 84, self.frame_depth),
                                                         n_actions=6, opt='adam', learning_rate=0.000102)})

        return config_dict

    def _build_for_double_dueling_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DoubleDuelingDQN'),
                            'double': True,
                            'model_architecture': DuelingConvNN(observation_shape=(84, 84, self.frame_depth),
                                                                n_actions=6, opt='adam', learning_rate=0.000102)})

        return config_dict

    def _build_for_random(self):
        return {'name': os.path.join(self.folder, 'RandomAgent'),
                'env_spec': self.env_spec}



from typing import Any, Tuple, Dict, Union
from unittest.mock import MagicMock

import numpy as np


class EnvFixture:
    action_space = MagicMock()
    reward_range = 1
    metadata = None

    def __init__(self, obs_shape: Tuple[int, int, int] = (250, 160)) -> None:
        self.obs_shape = obs_shape
        self._obs = np.ones(obs_shape)
        self._reward = 1.0
        self._stop_in: Union[None, int] = None

        self.observation_space = MagicMock()
        self.observation_space.shape = obs_shape

    def reset(self) -> np.ndarray:
        return self._obs

    def action_indicator(self, action: int) -> None:
        """Called each step, patch to monitor in tests."""
        pass

    def step(self, action: int) -> Tuple[np.ndarray, float, bool, Dict[Any, Any]]:
        self.action_indicator(action)

        if action == 1:
            # Trigger stop in a few turns
            if self._stop_in is None:
                self._stop_in = 3
            self._stop_in -= 1

        done = False
        info = {}
        reward = self._reward
        obs = self._obs

        if self._stop_in == 0:
            done = True
            obs += 1
            reward += 1

        return obs, reward, done, info



from dataclasses import dataclass
from typing import Any, Dict, Tuple, Union, Callable, Iterable

import numpy as np
from sklearn.linear_model import SGDRegressor

from reinforcement_learning_keras.agents.agent_base import AgentBase
from reinforcement_learning_keras.agents.components.helpers.env_builder import EnvBuilder
from reinforcement_learning_keras.agents.components.history.training_history import TrainingHistory
from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy
from reinforcement_learning_keras.enviroments.config_base import ConfigBase


@dataclass
class LinearQAgent(AgentBase):
    eps: EpsilonGreedy
    training_history: TrainingHistory
    env_spec: str = "CartPole-v0"
    env_wrappers: Iterable[Callable] = ()
    name: str = 'LinearQAgent'
    gamma: float = 0.99
    log_exemplar_space: bool = False
    final_reward: Union[float, None] = None

    def __post_init__(self, ) -> None:
        self.env_builder = EnvBuilder(self.env_spec, self.env_wrappers)
        self._build_model()

    def __getstate__(self) -> Dict[str, Any]:
        return self._pickle_compatible_getstate()

    def _build_model(self) -> None:
        """
        Build the models to estimate Q(a|s).

        Because this is linear regression with multiple outputs, use multiple models.
        """

        # Create SGDRegressor for each action in space and initialise by calling .partial_fit for the first time on
        # dummy data
        # Prep pipeline with scaler and rbfs paths

        mods = {a: SGDRegressor() for a in range(self.env.action_space.n)}
        for mod in mods.values():
            mod.partial_fit(self.transform(self.env.reset()), [0])

        self.mods = mods

    def transform(self, s: np.ndarray) -> np.ndarray:
        """Check shape. It's always single input with this agent."""
        if len(s.shape) == 1:
            s = s.reshape(1, -1)

        return s

    def partial_fit(self, s: np.ndarray, a: int, g: float) -> None:
        """
        Run partial fit for a single row of training data.

        :param s: The raw state observation.
        :param a: The action taken.
        :param g: The reward + discounted value of next state.
        """

        x = self.transform(s)
        self.mods[a].partial_fit(x, [g])

    def predict(self, s: np.ndarray) -> Dict[int, float]:
        """
        Given a single state observation, predict action values from each model.

        :param s: The raw state observation.
        :return: Dict containing action values, indexed by action id.
        """

        s = self.transform(s)

        return {a: float(mod.predict(s)) for a, mod in self.mods.items()}

    def get_best_action(self, s: np.ndarray) -> int:
        """
        Find the best action from the values predicted by each model.

        Remember preds is Dict[int, float]: This is essentially argmax on a dict.

        :param s: The raw state observation.
        :return: The action with the highest predicted value.
        """
        preds = self.predict(s)

        best_action_value = -np.inf
        best_action = 0
        for k, v in preds.items():
            if v > best_action_value:
                best_action_value = v
                best_action = k

        return best_action

    def get_action(self, s: np.ndarray, training: bool = False) -> int:
        """
        Get an action using epsilon greedy.

        Epsilon decays every time a random action is chosen.

        :param s: The raw state observation.
        :param training: Bool to indicate whether or not to use this experience to update the model. If False, just
                         returns best action.
        :return: The selected action.
        """
        action = self.eps.select(greedy_option=lambda: self.get_best_action(s),
                                 random_option=lambda: self.env.action_space.sample(),
                                 training=training)

        return action

    def update_model(self, s: np.ndarray, a: int, r: float, d: bool, s_: np.ndarray) -> None:
        """
        For a single step set, calculate discounted reward and update the appropriate action model.

        :param s: The raw state observation the action was selected for.
        :param a: The selected action.
        :param r: The reward for performing that action.
        :param d: Flag indicating if done.
        :param s_: The next state following the action.
        """
        # Update model
        if d:
            # Done flag is only true if env ends due to agent failure (not if max steps are reached). Punish.
            g = self.final_reward if self.final_reward is not None else 0
        else:
            # Calculate the reward for this step and the discounted max value of actions in the next state.
            g = r + self.gamma * np.max(list(self.predict(s_).values()))

        # Update the model s = x, g = y, and a is the model to update
        self.partial_fit(s, a, g)

    def _play_episode(self, max_episode_steps: int = 500,
                      training: bool = False, render: bool = True) -> Tuple[float, int]:
        """
        Play a single episode and return the total reward.

        :param max_episode_steps: Max steps before stopping, overrides any time limit set by Gym.
        :param training: Bool to indicate whether or not to use this experience to update the model.
        :param render: Bool to indicate whether or not to call env.render() each training step.
        :return: The total real reward for the episode.
        """
        self.env._max_episode_steps = max_episode_steps
        obs = self.env.reset()
        total_reward = 0
        for frame in range(max_episode_steps):
            action = self.get_action(obs, training=training)
            prev_obs = obs
            obs, reward, done, info = self.env.step(action)
            total_reward += reward

            if render:
                self.env.render()

            if training:
                self.update_model(s=prev_obs, a=action, r=reward, d=done, s_=obs)

            if done:
                break

        return total_reward, frame

    @classmethod
    def example(cls, config: ConfigBase, render: bool = True,
                n_episodes: int = 10, max_episode_steps: int = 500, update_every: int = 10) -> "AgentBase":
        """Create, train, and save agent for a given config."""
        config_dict = config.build()

        agent = cls(**config_dict)

        agent.train(verbose=True, render=render,
                    n_episodes=n_episodes, max_episode_steps=max_episode_steps, update_every=update_every)
        agent.save()

        return agent


if __name__ == "__main__":
    from reinforcement_learning_keras.enviroments.cart_pole import CartPoleConfig
    from reinforcement_learning_keras.enviroments import MountainCarConfig

    agent_cart_pole = LinearQAgent.example(CartPoleConfig(agent_type='linear_q', plot_during_training=True))
    agent_mountain_car = LinearQAgent.example(MountainCarConfig(agent_type='linear_q', plot_during_training=True))



from reinforcement_learning_keras.enviroments.atari.atari_default_config import AtariDefaultConfig


class PongConfig(AtariDefaultConfig):
    """Defines configs tweaks for Pong."""
    env_spec = 'PongNoFrameskip-v4'



import unittest
from unittest.mock import patch

from reinforcement_learning_keras.agents.agent_base import AgentBase
from reinforcement_learning_keras.agents.components.history.episode_report import EpisodeReport
from reinforcement_learning_keras.agents.random.random_agent import RandomAgent
from reinforcement_learning_keras.agents.random.random_model import RandomModel
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig


class TestRandomAgent(unittest.TestCase):
    """
    The random agent is basically a non-abstract version of AgentBase. It's used here to define the general test
    interface, as if it's a mocked AgentBase.

    Other agents can modify the private methods here to define how to arrange/assert for their specific cases.
    Acting is standard across agents, except for where path.object is required for multiple methods (see DQN as an
    example).
    """
    _sut = RandomAgent
    _config = CartPoleConfig(agent_type='random', plot_during_training=False)

    def _standard_set_up(self) -> None:
        self._n_step = 4
        self._n_episodes = 3
        # These are the calls during .play_episode (not .train)
        self._expected_model_update_during_training_episode: int = self._n_step
        self._expected_model_update_during_playing_episode: int = 0
        self._expected_play_episode_calls = self._n_episodes

    def _agent_specific_set_up(self) -> None:
        # This agent doesn't bother calling update_model as it doesn't have one.
        self._expected_model_update_during_training_episode: int = 0
        self._expected_model_update_after_training_episode: int = 0

    def setUp(self) -> None:
        self._standard_set_up()
        self._agent_specific_set_up()

    def _ready_agent(self) -> RandomAgent:
        return self._sut(**self._config.build())

    @staticmethod
    def _checkpoint_model(agent: AgentBase) -> None:
        """No model to checkpoint."""
        return None

    def _assert_model_unchanged(self, agent: AgentBase, checkpoint: None) -> None:
        """No model to compare, nothing to assert."""
        pass

    def _assert_buffer_changed(self, agent: AgentBase, checkpoint: None) -> None:
        """No buffer to change in RandomAgent."""
        pass

    def _assert_model_changed(self, agent: AgentBase, checkpoint: None) -> None:
        """No model to compare, nothing to assert."""
        pass

    def _assert_relevant_play_episode_change(self, agent: AgentBase, checkpoint: None) -> None:
        """This can differ between MC and TD agents. In MC case model might not be updated but buffer is."""
        self._assert_buffer_changed(agent, checkpoint)
        self._assert_model_changed(agent, checkpoint)

    def _assert_relevant_after_play_episode_change(self, agent: AgentBase, checkpoint: None) -> None:
        """This can differ between MC and TD agents. In MC case buffer might not be updated but model is."""
        self._assert_buffer_changed(agent, checkpoint)
        self._assert_model_changed(agent, checkpoint)

    def _assert_agent_unready(self, agent: AgentBase) -> None:
        """Nothing to unready in RandomAgent."""
        pass

    def _assert_agent_ready(self, agent: RandomAgent) -> None:
        self.assertIsNotNone(agent.env_builder._env)
        self.assertIsInstance(agent.model, RandomModel)

    def test_env_set_during_init(self) -> None:
        # Act
        agent = self._ready_agent()

        # Assert
        self.assertIsNotNone(agent.env_builder._env)

    def test_model_set_during_init(self) -> None:
        # Act
        agent = self._ready_agent()

        # Assert
        self._assert_agent_ready(agent)

    def test_history_set_during_init(self) -> None:
        # Act
        agent = self._ready_agent()

        # Assert
        self.assertIsNotNone(agent.training_history)

    def test_unready_detaches_env_and_models(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        agent.unready()

        # Assert
        self._assert_agent_unready(agent)

    def test_ready_restores_matching_object(self) -> None:
        # Arrange
        agent = self._ready_agent()
        checkpoint = self._checkpoint_model(agent)
        agent.unready()

        # Act
        agent.check_ready()

        # Assert
        self._assert_agent_ready(agent)
        self._assert_model_unchanged(agent, checkpoint)

    def test_play_episode_steps_returns_reward_when_not_training(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        reward = agent.play_episode(max_episode_steps=3, training=False, render=False)

        # Assert
        self.assertIsInstance(reward, EpisodeReport)

    def test_play_episode_steps_does_not_call_update_models_when_not_training(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        with patch.object(agent, 'update_model') as mocked_update_model:
            _ = agent.play_episode(max_episode_steps=self._n_step, training=False, render=False)

        # Assert
        self.assertEqual(self._expected_model_update_during_playing_episode, mocked_update_model.call_count)
        self.assertEqual(self._n_step, agent.env_builder._env._max_episode_steps)

    def test_play_episode_steps_does_not_update_models_when_not_training(self) -> None:
        # Arrange
        agent = self._ready_agent()
        checkpoint = self._checkpoint_model(agent)

        # Act
        _ = agent.play_episode(max_episode_steps=self._n_step, training=False, render=False)

        # Assert
        self._assert_model_unchanged(agent, checkpoint)

    def test_play_episode_steps_returns_reward_and_updates_model_when_training(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        reward = agent.play_episode(max_episode_steps=self._n_step, training=True, render=False)

        # Assert
        self.assertIsInstance(reward, EpisodeReport)

    def test_play_episode_steps_calls_update_models_as_expected_when_training(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        with patch.object(agent, 'update_model') as mocked_update_model:
            _ = agent.play_episode(max_episode_steps=self._n_step, training=True, render=False)

        # Assert
        self.assertEqual(self._expected_model_update_during_training_episode, mocked_update_model.call_count)

    def test_play_episode_steps_updates_models_as_expected_when_training(self) -> None:
        # Arrange
        agent = self._ready_agent()
        checkpoint = self._checkpoint_model(agent)

        # Act
        _ = agent.play_episode(max_episode_steps=self._n_step, training=True, render=False)

        # Assert
        self._assert_relevant_play_episode_change(agent, checkpoint)

    def test_train_runs_multiple_episodes(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        with patch.object(agent, 'play_episode') as mocked_play_episode, \
                patch.object(agent, '_after_episode_update') as after_ep_update:
            agent.train(n_episodes=self._n_episodes, max_episode_steps=self._n_step, render=False, checkpoint_every=0)

        # Assert
        self.assertEqual(self._n_episodes, len(agent.training_history.history))
        self.assertEqual(self._n_episodes, mocked_play_episode.call_count)
        self.assertEqual(self._n_episodes, after_ep_update.call_count)

    def test_train_calls_after_episode_updates_model_as_expected(self) -> None:
        # Arrange
        agent = self._ready_agent()
        checkpoint = self._checkpoint_model(agent)

        # Act
        agent.train(n_episodes=self._n_episodes, max_episode_steps=self._n_step, render=False, checkpoint_every=0)

        # Assert
        self._assert_relevant_after_play_episode_change(agent, checkpoint)

    def test_train_calls_after_episode_update_as_expected_with_delayed(self) -> None:
        # Arrange
        agent = self._ready_agent()

        # Act
        with patch.object(agent, 'play_episode') as mocked_play_episode, \
                patch.object(agent, '_after_episode_update') as after_ep_update:
            agent.train(n_episodes=self._n_episodes, max_episode_steps=self._n_step, render=False, checkpoint_every=0,
                        update_every=2)

        # Assert
        self.assertEqual(self._n_episodes, len(agent.training_history.history))
        self.assertEqual(self._n_episodes, mocked_play_episode.call_count)
        self.assertEqual(2, after_ep_update.call_count)



import os
from dataclasses import dataclass
from typing import List, Tuple, Union, Dict, Any, Callable, Iterable

import joblib
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K

from reinforcement_learning_keras.agents.agent_base import AgentBase
from reinforcement_learning_keras.agents.components.helpers.env_builder import EnvBuilder
from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.components.history.training_history import TrainingHistory
from reinforcement_learning_keras.agents.policy_gradient.loss import reinforce_loss
from reinforcement_learning_keras.enviroments.config_base import ConfigBase
from reinforcement_learning_keras.enviroments.model_base import ModelBase

tf.compat.v1.disable_eager_execution()


@dataclass
class ReinforceAgent(AgentBase):
    """
    Uses a simple replay buffer.

    Has 2 components:
      - _current_ : List of steps being collected for current episode
      - _buffer_ : Dict containing backlog of completed episodes not yet used for training model

    At the end of an episode, the current episode is moved to the backlog. This is cleared after updating model,
    which can occur less often.

    TODO: Move replay buffer to agents.components.replay_buffer.episodic_buffer
    """
    training_history: TrainingHistory
    model_architecture: ModelBase
    env_spec: str = "CartPole-v0"
    env_wrappers: Iterable[Callable] = ()
    name: str = 'REINFORCEAgent'
    alpha: float = 0.0001
    gamma: float = 0.99
    final_reward: Union[float, None] = None

    def __post_init__(self) -> None:
        self.env_builder = EnvBuilder(self.env_spec, self.env_wrappers)

        # Keep track of number of trained episodes, only used for IDing episodes in buffer.
        self._ep_tracker: int = -1
        self._fn = f"{self.name}_{self.env_spec}"
        self._build_model()
        self.clear_memory()
        self.ready = True

    def __getstate__(self) -> Dict[str, Any]:
        return self._pickle_compatible_getstate()

    def _save_model(self):
        if not os.path.exists(f"{self._fn}"):
            os.mkdir(f"{self._fn}")

        self._model.save(f"{self.name}_{self.env_spec}/model")

    def _load_model(self):

        self._model = keras.models.load_model(f"{self._fn}/model",
                                              custom_objects={'reinforce_loss': reinforce_loss})

    def unready(self) -> None:
        if self.ready:
            self._save_model()
            self._model = None
            keras.backend.clear_session()
            tf.compat.v1.reset_default_graph()
        super().unready()

    def check_ready(self):

        if not self.ready:
            self._load_model()

            super().check_ready()

    def _clear_current_episode(self) -> None:
        """Clear buffer for current episode."""
        self._current_states: List[np.ndarray] = []
        self._current_action_probs: List[np.ndarray] = []
        self._current_actions: List[int] = []
        self._current_rewards: List[float] = []

    def _clear_buffer_backlog(self) -> None:
        """Clear backlog of collected episodes not yet trained on."""
        self._buffer_states: Dict[int, List[np.ndarray]] = {}
        self._buffer_action_probs: Dict[int, List[np.ndarray]] = {}
        self._buffer_actions: Dict[int, np.ndarray] = {}
        self._buffer_rewards: Dict[int, np.ndarray] = {}
        self._buffer_discounted_rewards: Dict[int, np.ndarray] = {}

    def _move_current_episode_to_backlog(self, episode: int):
        """Move current episode to backlog, calc discounted rewards, and clear."""
        self._buffer_states[episode] = self._current_states
        self._buffer_action_probs[episode] = self._current_action_probs
        self._buffer_actions[episode] = np.array(self._current_actions)
        self._buffer_rewards[episode] = np.array(self._current_rewards)
        self._buffer_discounted_rewards[episode] = self._calc_discounted_rewards(self._current_rewards)
        self._clear_current_episode()

    def clear_memory(self) -> None:
        """Clear current episode and backlog buffers."""
        self._clear_current_episode()
        self._clear_buffer_backlog()

    def _build_model(self) -> None:
        """State -> model -> action probs"""
        self._model = self.model_architecture.compile(model_name='action_model', loss=reinforce_loss)

    def transform(self, s: Union[List[np.ndarray], np.ndarray]) -> np.ndarray:
        """No transforming of state here, just stacking and dimension checking."""
        if len(s.shape) < len(self._model.input.shape):
            s = np.expand_dims(s, 0)

        return s

    def get_action(self, s: np.ndarray, training=None) -> Tuple[np.ndarray, int]:
        """
        Use the current policy to select an action from a single state observation.

        Sample actions using the probabilities provided by the action model.
        """
        actions_probs = self._model.predict(self.transform(s)).squeeze()
        return actions_probs, np.random.choice(range(self.env.action_space.n),
                                               p=actions_probs)

    def update_experience(self, s: np.ndarray, a: int, r: float, a_p: np.ndarray) -> None:
        """
        Add step of experience to the buffer.

        :param s: State
        :param a: Action
        :param r: Reward
        :param a_p: Action probabilities
        """
        self._current_states.append(s)
        self._current_action_probs.append(a_p)
        self._current_actions.append(a)
        self._current_rewards.append(r)

    def _calc_discounted_rewards(self, rr: List[float]) -> np.ndarray:
        """Calculate discounted rewards for a whole episode and normalise."""

        # Full episode returns
        disc_rr = np.zeros_like(rr)
        cumulative_reward = 0
        for t in reversed(range(0, disc_rr.size)):
            cumulative_reward = cumulative_reward * self.gamma + rr[t]
            disc_rr[t] = cumulative_reward

        # Normalise
        disc_rr_mean = np.mean(disc_rr)
        disc_rr_std = np.std(disc_rr) + 1e-9
        disc_rr_norm = (disc_rr - disc_rr_mean) / disc_rr_std

        return np.vstack(disc_rr_norm)

    @staticmethod
    def _flatten_list(nested_list: List[List[Any]]) -> List[Any]:
        return [item for sublist in nested_list for item in sublist]

    def update_model(self) -> None:
        # Stack all available episodes

        states = np.concatenate(list(self._buffer_states.values()))
        disc_rewards = np.concatenate(list(self._buffer_discounted_rewards.values()))
        actions = np.concatenate(list(self._buffer_actions.values()))
        action_probs = np.vstack(list(self._buffer_action_probs.values()))

        # One hot actions
        actions_oh = K.one_hot(actions,
                               num_classes=self.env.action_space.n)

        # Calculate prob updates
        dlogps = (actions_oh - action_probs) * disc_rewards
        y = action_probs + self.alpha * dlogps

        # Train
        x = self.transform(states)
        self._model.train_on_batch(x, y)

    def _play_episode(self, max_episode_steps: int = 500,
                      training: bool = False, render: bool = True) -> Tuple[float, int]:
        self.env._max_episode_steps = max_episode_steps

        obs = self.env.reset()
        total_reward = 0

        for frame in range(max_episode_steps):
            action_probs, action = self.get_action(obs)
            prev_obs = obs
            obs, reward, done, _ = self.env.step(action)
            total_reward += reward

            if render:
                self.env.render()

            if training:
                self.update_experience(s=prev_obs, a=action, r=reward, a_p=action_probs)

            if done:
                break

        if training:
            # Only keep episode buffer if actually training
            self._ep_tracker += 1
            self._move_current_episode_to_backlog(self._ep_tracker)

        return total_reward, frame

    def _after_episode_update(self) -> None:
        """Monte-Carlo update of policy model is updated (ie. after each full episode, or more)"""
        self.update_model()
        self.clear_memory()

    def save(self) -> None:
        # No need to unready, this uses __getstate__
        if not os.path.exists(f"{self._fn}"):
            os.mkdir(f"{self._fn}")

        self.unready()
        joblib.dump(self, f"{self.name}_{self.env_spec}/agent.joblib")
        self.check_ready()

    @classmethod
    def load(cls, fn: str) -> "ReinforceAgent":
        new_agent = joblib.load(f"{fn}/agent.joblib")
        new_agent.check_ready()

        return new_agent

    @classmethod
    def example(cls, config: ConfigBase, render: bool = True,
                n_episodes: int = 500, max_episode_steps: int = 500, update_every: int = 1) -> "ReinforceAgent":
        """Create, train, and save agent for a given config."""
        VirtualGPU(config.gpu_memory)
        config_dict = config.build()

        agent = cls(**config_dict)

        agent.train(verbose=True, render=render,
                    n_episodes=n_episodes, max_episode_steps=max_episode_steps, update_every=update_every)
        agent.save()

        return agent


if __name__ == "__main__":
    from reinforcement_learning_keras.enviroments.cart_pole import CartPoleConfig

    agent_cart_pole = ReinforceAgent.example(CartPoleConfig(agent_type='reinforce', plot_during_training=True),
                                             render=False)



from reinforcement_learning_keras.enviroments.atari.space_invaders.space_invaders_config import SpaceInvadersConfig
from tests.integration.environments.atari.pong.test_pong_config import TestPongStackEnvironment, TestPongDiffEnvironment


class TestSpaceInvadersStackEnvironment(TestPongStackEnvironment):
    _sut = SpaceInvadersConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (84, 84, 3)
    _n_steps = 20
    _show = False


class TestSpaceInvadersDiffEnv(TestPongDiffEnvironment):
    _sut = SpaceInvadersConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (84, 84, 1)
    _n_steps = 20
    _show = False


del TestPongStackEnvironment
del TestPongDiffEnvironment



from reinforcement_learning_keras.enviroments.doom.doom_default_config import DoomDefaultConfig


class VizDoomPredictPositionConfig(DoomDefaultConfig):
    """Defines specific config for this env."""
    env_spec = 'VizdoomPredictPosition-v0'
    n_actions = 3
    res_scale = 0.2
    target_obs_shape = (90, 160)



from reinforcement_learning_keras.enviroments.doom.doom_default_config import DoomDefaultConfig


class VizDoomMyWayHomeConfig(DoomDefaultConfig):
    """Defines specific config for this env."""
    env_spec = 'VizdoomMyWayHome-v0'
    n_actions = 3
    res_scale = 0.2


import warnings

try:
    from reinforcement_learning_keras.enviroments.doom.vizdoom_basic_config import VizDoomBasicConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_corridor_config import VizDoomCorridorConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_deathmatch_config import VizDoomDeathmatchConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_defend_center_config import VizDoomDefendCenterConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_defend_line_config import VizDoomDefendLineConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_health_gathering_config import \
        VizDoomHealthGatheringConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_health_gathering_supreme import \
        VizDoomHealthGatheringSupremeConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_my_way_home_config import VizDoomMyWayHomeConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_predict_position_config import \
        VizDoomPredictPositionConfig
    from reinforcement_learning_keras.enviroments.doom.vizdoom_take_cover_config import VizDoomTakeCoverConfig

    AVAILABLE = True
except ImportError:
    warnings.warn('vizdoomgym not installed.')
    AVAILABLE = False



import os
from typing import Any, Dict

from reinforcement_learning_keras.agents.components.replay_buffers.continuous_buffer import ContinuousBuffer
from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy
from reinforcement_learning_keras.enviroments.cart_pole.models.small_dueling_nn import SmallDuelingNN
from reinforcement_learning_keras.enviroments.config_base import ConfigBase
from reinforcement_learning_keras.enviroments.mountain_car.models.small_nn import SmallNN


class MountainCarConfig(ConfigBase):
    """Defines config for mountain_car"""
    env_spec = 'MountainCar-v0'
    supported_agents = ('linear_q', 'dueling_dqn', 'dqn', 'random')
    gpu_memory = 128

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _default_training_history_kwargs(self) -> Dict[str, Any]:
        return {"plotting_on": self.plot_during_training, "plot_every": 200, "rolling_average": 12}

    def _build_for_linear_q(self) -> Dict[str, Any]:
        return {'name': os.path.join(self.folder, 'LinearQAgent'),
                'env_spec': self.env_spec,
                'final_reward': 500,
                'gamma': 0.99,
                'log_exemplar_space': False,
                'eps': EpsilonGreedy(eps_initial=0.3, eps_min=0.005)}

    def _build_for_dqn(self) -> Dict[str, Any]:
        """This isn't tuned."""
        return {'name': os.path.join(self.folder, 'DeepQAgent'),
                'env_spec': self.env_spec,
                'model_architecture': SmallNN(observation_shape=(2,), n_actions=3, output_activation=None,
                                              opt='adam', learning_rate=0.001),
                'gamma': 0.99,
                'final_reward': 650,
                'replay_buffer_samples': 32,
                'eps': EpsilonGreedy(eps_initial=0.1, decay=0.002, eps_min=0.002),
                'replay_buffer': ContinuousBuffer(buffer_size=200)}

    def _build_for_dueling_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DuelingDQN'),
                            'model_architecture': SmallDuelingNN(observation_shape=(2,), n_actions=3,
                                                                 opt='adam', learning_rate=0.001)})

        return config_dict

    def _build_for_random(self):
        return {'name': os.path.join(self.folder, 'RandomAgent'),
                'env_spec': self.env_spec}



import numpy as np


class RandomModel:
    def __init__(self, n_actions: int):
        self.n_actions = n_actions

    def predict(self) -> int:
        return int(np.random.randint(0, self.n_actions, 1))



import copy
from typing import List

import numpy as np
from numpy.testing import assert_array_almost_equal

from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.policy_gradient.reinforce_agent import ReinforceAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from tests.unit.agents.random.test_random_agent import TestRandomAgent


class TestReinforceAgent(TestRandomAgent):
    _sut = ReinforceAgent
    _config = CartPoleConfig(agent_type='reinforce', plot_during_training=False)

    def _agent_specific_set_up(self):
        VirtualGPU(256)

        # For this agent, this call is made once at the end of the episode not on every step
        self._expected_model_update_during_training_episode: int = 0
        # This is inside the play episode function, which is still 0 here as it's called in .train
        self._expected_model_update_after_playing_episode: int = 0

    @staticmethod
    def _checkpoint_model(agent: ReinforceAgent) -> List[np.ndarray]:
        """Get coefs from each model"""
        return copy.deepcopy(agent._model.get_weights())

    def _assert_model_unchanged(self, agent: ReinforceAgent, checkpoint: List[np.ndarray]):
        model_weights = agent._model.get_weights()
        for w in range(len(model_weights)):
            assert_array_almost_equal(model_weights[w], checkpoint[w])

    def _assert_relevant_play_episode_change(self, agent: ReinforceAgent, checkpoint: List[np.ndarray]) -> None:
        self._assert_buffer_changed(agent, checkpoint)

    def _assert_relevant_after_play_episode_change(self, agent: ReinforceAgent, checkpoint: List[np.ndarray]) -> None:
        self._assert_model_changed(agent, checkpoint)

    def _assert_buffer_changed(self, agent: ReinforceAgent, checkpoint: List[np.ndarray]):
        # TODO
        pass

    def _assert_model_changed(self, agent: ReinforceAgent, checkpoint: List[np.ndarray]) -> None:
        model_weights = agent._model.get_weights()
        for w in range(len(checkpoint)):
            self.assertFalse(np.all(model_weights[w].round(6) == checkpoint[w].round(6)))

    def _assert_agent_unready(self, agent: ReinforceAgent) -> None:
        self.assertIsNone(agent._model)
        self.assertFalse(agent.ready)

    def _assert_agent_ready(self, agent: ReinforceAgent) -> None:
        self.assertIsNotNone(agent.env_builder._env)
        self.assertIsNotNone(agent._model)
        self.assertTrue(agent.ready)

    def test_train_calls_after_episode_updates_model_as_expected(self) -> None:
        # Arrange
        agent = self._ready_agent()
        checkpoint = self._checkpoint_model(agent)

        # Act
        agent.train(n_episodes=self._n_episodes, max_episode_steps=self._n_step, render=False, checkpoint_every=0)

        # Assert
        self._assert_relevant_after_play_episode_change(agent, checkpoint)


# Prevent parent test from being collected again
del TestRandomAgent



import gym
import numpy as np
from sklearn.kernel_approximation import RBFSampler
from sklearn.pipeline import FeatureUnion


class RBFSWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)

        # Sample observations from env and fit pipeline
        obs = np.array([self.env.observation_space.sample() for _ in range(10000)])
        rbfs = FeatureUnion([('rbf0', RBFSampler(gamma=100, n_components=60)),
                             ('rbf1', RBFSampler(gamma=1, n_components=60)),
                             ('rbf2', RBFSampler(gamma=0.02, n_components=60))])
        trans_obs = rbfs.fit_transform(obs)
        self.rbfs = rbfs

        # New env obs space shape
        self.observation_space = gym.spaces.Box(low=trans_obs.min() * 2,
                                                high=trans_obs.max() * 2,
                                                shape=(180,))

    def observation(self, obs: np.ndarray) -> np.ndarray:
        return self.rbfs.transform(obs.reshape(1, -1))



from dataclasses import dataclass
import warnings


@dataclass
class VirtualGPU:
    """
    gpu_memory_limit: Max memory in MB for  virtual device. Setting LOWER than total available this can help
                      avoid out of memory errors on some set ups when TF tries to allocate too much memory
                      (seems to be a bug).
    gpu_device_id: Integer device identifier for the real GPU the virtual GPU should use.
    """
    gpu_memory_limit: int = 512
    gpu_device_id: int = 0

    def __post_init__(self):
        self.physical_device = None
        self.virtual_device = None
        self.on = self._set_tf()

    def _set_tf(self) -> bool:
        """
        Helper function for training on tf. Reduces GPU memory footprint for keras/tf models.

        Creates a virtual device on the request GPU with limited memory. Will fail gracefully if GPU isn't available.

        :return: Bool indicating if TF appears to be running on GPU. Can be used, for example, to avoid using
                 multiprocessing in the caller when running on GPU. This will likely result in an exception, but may
                 result in hanging forever, so probably best avoided.
        """
        import tensorflow as tf

        try:
            self.physical_device = tf.config.experimental.list_physical_devices('GPU')[self.gpu_device_id]
            virtual_device = tf.config.experimental.VirtualDeviceConfiguration(memory_limit=self.gpu_memory_limit)
        except (IndexError, AttributeError) as e:
            # IndexError: Assuming using GPU but indexed device not found.
            # AAttributeError: Assuming no GPU.
            warnings.warn(f"Not using GPU due to: {e}")
            return False

        # First check a virtual device hasn't already been set. If it has, we don't want to try and set a new one.
        # - If the device has not been used before, it will be replaced and no error is raised
        # - If the device has been used before it will be initialised and will be immutable, raising a RuntimeError
        #   on set_virtual_device_configuration call.
        # Aim here is to make behaviour more predictable. This allows multiple models to run in the same session, as
        # long as the max memory required is set at the start. The alternative is it fixing to the first used model's
        # memory requirement (for example in a session that calls VirtualGPU again in agent.example()).
        existing_device = tf.config.experimental.get_virtual_device_configuration(self.physical_device)
        if existing_device is not None:
            warnings.warn(f"A virtual GPU with {existing_device[0].memory_limit} MB memory already exists, "
                          f"using this rather than creating another with the requested {self.gpu_memory_limit} MB."
                          f"Good luck.")
            self.virtual_device = existing_device[0]
            self.gpu_memory_limit = existing_device[0].memory_limit

        else:
            # GPU available and no existing virtual device. Create a new one.
            tf.config.experimental.set_virtual_device_configuration(self.physical_device, [virtual_device])

        return True



from dataclasses import dataclass
from typing import Callable, Any

import numpy as np


@dataclass
class EpsilonGreedy:
    """Handles epsilon-greedy action selection, decay of epsilon during training."""
    eps_initial: float = 0.2
    decay: float = 0.0001
    decay_schedule: str = 'compound'
    eps_min: float = 0.01
    state = None

    def __post_init__(self) -> None:
        self.eps_current = self.eps_initial

        valid_decay = ('linear', 'compound')
        if self.decay_schedule.lower() not in valid_decay:
            raise ValueError(f"Invalid decay schedule {self.decay_schedule}. Pick from {valid_decay}.")

        self._set_random_state()

    def _set_random_state(self) -> None:
        self._state = np.random.RandomState(self.state)

    def _linear_decay(self) -> float:
        return self.eps_current - self.decay

    def _compound_decay(self) -> float:
        return self.eps_current - self.eps_current * self.decay

    def _decay(self):
        new_eps = np.nan
        if self.decay_schedule.lower() == 'linear':
            new_eps = self._linear_decay()

        if self.decay_schedule.lower() == 'compound':
            new_eps = self._compound_decay()

        return max(self.eps_min, new_eps)

    def select(self, greedy_option: Callable, random_option: Callable,
               training: bool = False) -> Any:
        """
        Apply epsilon greedy selection.

        If training, decay epsilon, and return selected option. If not training, just return greedy_option.

        Use of lambdas is to avoid unnecessarily picking between two pre-computed options.

        :param greedy_option: Function to evaluate if random option is NOT picked.
        :param random_option: Function to evaluate if random option IS picked.
        :param training: Bool indicating if call is during training and to use epsilon greedy and decay.
        :return: Evaluated selected option.
        """
        if training:
            self.eps_current = self._decay()
            if self._state.random() < self.eps_current:
                return random_option()

        return greedy_option()

    @classmethod
    def future_value(cls, eps: float, decay: float, steps: int, decay_schedule: str) -> float:
        """
        Calculate what eps will be after a number of steps.

        Decay is either decay% per step (like compound interest), or linear with constant decay amount.

        :param eps: Current/initial epsilon.
        :param decay: Decay rate per step.
        :param steps: Number of steps (usually training frames, rather than whole episodes)
        :param decay_schedule: 'linear' or 'compound'.
        :return:
        """
        if decay_schedule.lower() == 'compound':
            return eps * (1 - decay) ** steps

        if decay_schedule.lower() == 'linear':
            return eps - decay * steps



import unittest
from unittest.mock import MagicMock

import reinforcement_learning_keras.enviroments.doom as doom
from tests.integration.environments.atari.pong.test_pong_config import TestPongStackEnvironment, TestPongDiffEnvironment

if doom.AVAILABLE:
    run_tests = True
else:
    run_tests = False
    doom = MagicMock()


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomBasicConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomCorridorConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomDeathmatchConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomDefendCenterConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomDefendLineConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomHealthGatheringConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomHealthGatheringSupremeConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomMyWayHomeConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomPredictPositionConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (90, 160, 3)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongStackEnvironment):
    _sut = doom.VizDoomTakeCoverConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 3)
    _show = False


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomStackEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomBasicConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomCorridorConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomDeathmatchConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomDefendCenterConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomDefendLineConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomHealthGatheringConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomHealthGatheringSupremeConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomMyWayHomeConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomPredictPositionConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (90, 160, 1)


@unittest.skipUnless(run_tests, reason='ViZDoomGym not installed')
class TestDoomDiffEnvironments(TestPongDiffEnvironment):
    _sut = doom.VizDoomTakeCoverConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (96, 128, 1)
    _show = False


del TestPongStackEnvironment
del TestPongDiffEnvironment



import unittest
from unittest.mock import call, Mock

from reinforcement_learning_keras.enviroments.atari.environment_processing.fire_start_wrapper import FireStartWrapper
from tests.unit.environments.pong.environment_processing.env_fixture import EnvFixture


class TestFireStartWrapper(unittest.TestCase):
    _sut = FireStartWrapper
    _env_fixture = EnvFixture

    def setUp(self):
        self._env = self._env_fixture()

    def test_fire_is_first_action(self):
        # Arrange
        env = self._sut(self._env, fire_action_id=-10)
        env.env.action_indicator = Mock()

        # Act
        _ = env.reset()
        obs, reward, done, _ = env.step(action=0)

        # Assert
        env.env.action_indicator.assert_has_calls(calls=[call(-10), call(0)])

    def test_reset_returns_obs_of_expected_shape(self):
        # Arrange
        env = self._sut(self._env, fire_action_id=-10)

        # Act
        obs = env.reset()

        self.assertEqual(self._env.obs_shape, obs.shape)

    def test_step_returns_obs_of_expected_shape(self):
        # Arrange
        env = self._sut(self._env, fire_action_id=-10)
        _ = env.reset()

        # Act
        obs, reward, done, _ = env.step(action=0)

        self.assertEqual(self._env.obs_shape, obs.shape)



from typing import Type

import cv2
import gym
import numpy as np


class ImageProcessWrapper(gym.ObservationWrapper):
    """Scale image by given factor and make greyscale."""

    def __init__(self, env: gym.Env, scale: float = 0.4, dtype: Type = np.float32) -> None:
        super().__init__(env)
        self.dtype = dtype
        self.scale = scale

        self._new_size = (int(env.observation_space.shape[0] * scale), int(env.observation_space.shape[1] * scale))
        # New env obs space shape
        self.observation_space = gym.spaces.Box(low=0, high=1, shape=self._new_size, dtype=self.dtype)

    def observation(self, obs: np.ndarray) -> np.ndarray:
        return self.process(obs)

    def process(self, frame: np.ndarray) -> np.ndarray:
        gs = np.sum(frame, axis=2) / 3
        gs_resized = cv2.resize(gs, (self._new_size[1], self._new_size[0]), interpolation=cv2.INTER_AREA)
        gs_norm = gs_resized.astype(self.dtype) / 255.0

        return gs_norm



from functools import partial

from reinforcement_learning_keras.enviroments.atari.atari_default_config import AtariDefaultConfig
from reinforcement_learning_keras.enviroments.atari.environment_processing.fire_start_wrapper import FireStartWrapper
from reinforcement_learning_keras.enviroments.atari.environment_processing.frame_buffer_wrapper import \
    FrameBufferWrapper
from reinforcement_learning_keras.enviroments.atari.environment_processing.image_process_wrapper import \
    ImageProcessWrapper
from reinforcement_learning_keras.enviroments.atari.environment_processing.max_and_skip_wrapper import MaxAndSkipWrapper


class SpaceInvadersConfig(AtariDefaultConfig):
    """Defines configs tweaks for Space Invaders."""
    env_spec = 'SpaceInvadersNoFrameskip-v0'

    _wrappers_stack = (partial(MaxAndSkipWrapper, frame_buffer_length=4),
                       ImageProcessWrapper,
                       FireStartWrapper,
                       FrameBufferWrapper)
    _wrappers_diff = (partial(MaxAndSkipWrapper, frame_buffer_length=4),
                      ImageProcessWrapper,
                      FireStartWrapper,
                      partial(FrameBufferWrapper, buffer_length=2, buffer_function='diff'))



import collections
from typing import Any, Dict, Tuple, Callable

import gym
import numpy as np


class MaxAndSkipWrapper(gym.Wrapper):
    """Pools frames across steps, returning max, and repeat action for a number of frames."""

    def __init__(self, env: gym.Env,
                 frame_buffer_length: int = 2,
                 n_action_frames: int = 4,
                 frame_buffer_agg_f: Callable = np.max) -> None:
        """

        :param env: Gym environment to wrap, "inner environment".
        :param frame_buffer_length: Max number of frames to collect. FIFO buffer, contains most recent frames only.
        :param n_action_frames: Number of frames to repeat action for. Can be longer than frame buffer.
        :param frame_buffer_agg_f: Function used to aggregate frames in frame buffer to create observation.
                                   Default np.max.
        """
        super().__init__(env)
        self._n_action_frames = n_action_frames
        self._frame_buffer_length = frame_buffer_length
        self._frame_buffer_agg_f = frame_buffer_agg_f
        self._prepare_frame_buffer()

    def _prepare_frame_buffer(self) -> None:
        self._frame_buffer = collections.deque(maxlen=self._frame_buffer_length)

    def _aggregate_buffer_frames(self, ) -> np.ndarray:
        return self._frame_buffer_agg_f(np.stack(self._frame_buffer), axis=0)

    def step(self, action: int) -> Tuple[np.ndarray, float, bool, Dict[Any, Any]]:
        """
        Each step call iterates inner env 4 times, max of these is returned.

        Same action is applied for each step, and reward accumulated. If done, loop breaks and returns outputs
        based on shorter pool.

        :param action: Int id of action to perform.
        """
        total_reward = 0.0
        done = False
        info = {}
        for _ in range(self._n_action_frames):
            obs, reward, done, info = self.env.step(action)
            self._frame_buffer.append(obs)
            total_reward += reward
            if done:
                break

        agg_frame = self._aggregate_buffer_frames()

        return agg_frame, total_reward, done, info

    def reset(self) -> np.ndarray:
        """
        Additionally clears frame buffer

        :return: Observation from inner_env.reset() call.
        """
        self._frame_buffer.clear()
        obs = self.env.reset()
        self._frame_buffer.append(obs)

        return obs



"""General script to load a trained agent and play an episode with a monitor wrapper."""

from gym import wrappers

from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent


def load_and_play(agent_name: str, env_spec: str, output_dir: str = None) -> None:
    """Load agent, get it's env, wrap, play."""

    agent = DeepQAgent.load(f"{agent_name}_{env_spec}")

    if output_dir is None:
        output_dir = f'{agent.name}_{env_spec}_monitor_dir'

    # Replace agents environment with a wrapped version of the same env, then play episode. Render doesn't have to be
    # True here to save output to disk.
    agent.env_builder.set_env(wrappers.Monitor(agent.env, output_dir, force=True))
    agent.play_episode(training=False, render=True, max_episode_steps=10000)
    agent.env.close()


if __name__ == "__main__":
    load_and_play(agent_name="DeepQAgent", env_spec="VizdoomBasic-v0")




from typing import Tuple

from tensorflow import keras

from reinforcement_learning_keras.enviroments.model_base import ModelBase


class SmallNN(ModelBase):

    def _model_architecture(self) -> Tuple[keras.layers.Layer, keras.layers.Layer]:
        n_units = 16 * self.unit_scale

        state_input = keras.layers.Input(name='input', shape=self.observation_shape)
        fc1 = keras.layers.Dense(units=int(n_units), name='fc1', activation='relu')(state_input)
        fc2 = keras.layers.Dense(units=int(n_units / 2), name='fc2', activation='relu')(fc1)
        action_output = keras.layers.Dense(units=self.n_actions, name='output', activation=self.output_activation)(fc2)

        return state_input, action_output



from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.enviroments.atari.pong.pong_config import PongConfig

if __name__ == "__main__":
    agent = DeepQAgent.example(config=PongConfig(agent_type='double_dueling_dqn', mode='stack'), max_episode_steps=10000,
                               render=True, update_every=1, checkpoint_every=0)
    agent.save()



import collections
from dataclasses import dataclass
from typing import Tuple, Any, Iterable, List

import numpy as np

from reinforcement_learning_keras.agents.components.replay_buffers.replay_buffer_base import ReplayBufferBase


@dataclass
class ContinuousBuffer(ReplayBufferBase):
    buffer_size: int = 50

    def __post_init__(self) -> None:
        self._state_queue = collections.deque(maxlen=self.buffer_size)
        self._other_queue = collections.deque(maxlen=self.buffer_size)

        self.queue = collections.deque(maxlen=self.buffer_size)

    def __len__(self) -> int:
        return self.n if self.n > 0 else 0

    @property
    def full(self) -> bool:
        return len(self._state_queue) == self.buffer_size

    @property
    def n(self) -> int:
        return len(self._state_queue) - 1

    def append(self, items: Tuple[Any, int, float, bool]) -> None:
        """
        :param items: Tuple containing (s, a, r, d).
        """
        self._state_queue.append(items[0])
        self._other_queue.append(items[1::])

    def get_batch(self, idxs: Iterable[int]) -> Tuple[List[np.ndarray], List[np.ndarray],
                                                      List[float], List[bool], List[np.ndarray]]:
        ss = [self._state_queue[i] for i in idxs]
        ss_ = [self._state_queue[i + 1] for i in idxs]

        ard = [self._other_queue[i] for i in idxs]
        aa = [a for (a, _, _) in ard]
        rr = [r for (_, r, _) in ard]
        dd = [d for (_, _, d) in ard]

        return ss, aa, rr, dd, ss_

    def sample_batch(self, n: int) -> Tuple[List[np.ndarray], List[np.ndarray],
                                            List[float], List[bool], List[np.ndarray]]:
        if n > self.n:
            raise ValueError

        idxs = np.random.randint(0, self.n, n)
        return self.get_batch(idxs)



import gc
import tempfile
import unittest
from unittest.mock import MagicMock

import tensorflow as tf

from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.components.replay_buffers.continuous_buffer import ContinuousBuffer
from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy
from reinforcement_learning_keras.enviroments.atari.pong.pong_config import PongConfig
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from reinforcement_learning_keras.enviroments.mountain_car.mountain_car_config import MountainCarConfig


class TestDeepQAgent(unittest.TestCase):
    _sut = DeepQAgent
    _fn = 'test_dqn_save.agents'
    _gpu = VirtualGPU(1024)

    def setUp(self) -> None:
        self._tmp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        tf.keras.backend.clear_session()
        tf.compat.v1.reset_default_graph()
        gc.collect()
        self._tmp_dir.cleanup()

    @staticmethod
    def _build_mock_config(base_config: PongConfig) -> MagicMock:
        config = base_config.build()
        config['eps'] = EpsilonGreedy(eps_initial=0.5, decay=0.0001, eps_min=0.01, decay_schedule='linear')
        config['replay_buffer'] = ContinuousBuffer(buffer_size=10)
        config['replay_buffer_samples'] = 2
        mock_config = MagicMock()
        mock_config.gpu_memory = 2048
        mock_config.build.return_value = config

        return mock_config

    def test_saving_and_reloading_creates_identical_object(self):
        # Arrange
        agent = self._sut(**CartPoleConfig(agent_type='dqn', plot_during_training=False,
                                           folder=self._tmp_dir.name).build())
        agent.train(verbose=True, render=False, n_episodes=2)

        # Act
        agent.save()
        agent_2 = self._sut.load(f"{agent.name}_{agent.env_spec}")
        agent_2.check_ready()

        # Assert
        self.assertEqual(agent, agent_2)

    def test_dqn_cart_pole_example(self):
        # Arrange
        config = CartPoleConfig(agent_type='dqn', plot_during_training=False,
                                folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, n_episodes=10)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_dqn_mountain_car_example(self):
        # Arrange
        config = MountainCarConfig(agent_type='dqn', plot_during_training=False,
                                   folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, max_episode_steps=50, n_episodes=10)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_dqn_pong_diff_example(self):
        # Arrange
        mock_config = self._build_mock_config(PongConfig(agent_type='dqn', mode='diff', plot_during_training=False,
                                                         folder=self._tmp_dir.name))

        # Act
        # Needs to run for long enough to fill replay buffer
        agent = self._sut.example(mock_config, render=False, max_episode_steps=20, n_episodes=3)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_dqn_pong_stack_example(self):
        # Arrange
        mock_config = self._build_mock_config(PongConfig(agent_type='dqn', mode='stack', plot_during_training=False,
                                                         folder=self._tmp_dir.name))

        # Act
        # Needs to run for long enough to fill replay buffer
        agent = self._sut.example(mock_config, render=False, max_episode_steps=20, n_episodes=3)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_dueling_dqn_cart_pole_example(self):
        # Arrange
        config = CartPoleConfig(agent_type='dueling_dqn', plot_during_training=False,
                                folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, n_episodes=18)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_dueling_dqn_mountain_car_example(self):
        # Arrange
        config = MountainCarConfig(agent_type='dueling_dqn', plot_during_training=False,
                                   folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, max_episode_steps=100, n_episodes=18)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_double_dqn_cart_pole_example(self):
        # Arrange
        config = CartPoleConfig(agent_type='double_dqn', plot_during_training=False,
                                folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, n_episodes=16)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_double_dueling_dqn_cart_pole_example(self):
        # Arrange
        config = CartPoleConfig(agent_type='double_dueling_dqn', plot_during_training=False,
                                folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, n_episodes=20)

        # Assert
        self.assertIsInstance(agent, self._sut)



from typing import Type

import cv2
import gym
import numpy as np


class ImageProcessWrapper(gym.ObservationWrapper):
    """
    Convert frames from (210, 160, 3) or (250, 160, 3) to (84, 84, 1).
    Scales from 0 -> 255 to 0 -> 1.

    Based on ProcessFrame84, ScaledFloatFrame, ImageToPyTorch wrappers:
    https://github.com/PacktPublishing/Deep-Reinforcement-Learning-Hands-On/blob/master/Chapter06/lib/wrappers.py
    """

    def __init__(self, env: gym.Env, dtype: Type = np.float32) -> None:
        super().__init__(env)
        self.dtype = dtype
        # New env obs space shape
        self.observation_space = gym.spaces.Box(low=0, high=1, shape=(84, 84, 1),
                                                dtype=self.dtype)

    def observation(self, obs: np.ndarray) -> np.ndarray:
        return self.process(obs)

    def process(self, frame: np.ndarray) -> np.ndarray:
        if frame.size == 210 * 160 * 3:
            img = np.reshape(frame, [210, 160, 3]).astype(self.dtype)
        elif frame.size == 250 * 160 * 3:
            img = np.reshape(frame, [250, 160, 3]).astype(self.dtype)
        else:
            raise ValueError("Unknown resolution.")
        img = img[:, :, 0] * 0.299 + img[:, :, 1] * 0.587 + img[:, :, 2] * 0.114
        resized_screen = cv2.resize(img, (84, 110), interpolation=cv2.INTER_AREA)
        x_t = resized_screen[18:102, :]
        x_t = np.reshape(x_t, [84, 84, 1]).astype(self.dtype) / 255.0

        return x_t.squeeze()



import numpy as np

from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.enviroments.doom.vizdoom_basic_config import VizDoomBasicConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_corridor_config import VizDoomCorridorConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_deathmatch_config import VizDoomDeathMatchConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_defend_center_config import VizDoomDefendCenterConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_defend_line_config import VizDoomDefendLineConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_health_gathering_config import VizDoomHealthGatheringConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_health_gathering_supreme import \
    VizDoomHealthGatheringSupremeConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_my_way_home_config import VizDoomMyWayHomeConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_predict_position_config import VizDoomPredictPositionConfig
from reinforcement_learning_keras.enviroments.doom.vizdoom_take_cover_config import VizDoomTakeCoverConfig

if __name__ == "__main__":
    selected_config = VizDoomCorridorConfig

    available_configs = [VizDoomBasicConfig, VizDoomCorridorConfig, VizDoomDefendCenterConfig, VizDoomDefendLineConfig,
                         VizDoomHealthGatheringConfig, VizDoomMyWayHomeConfig, VizDoomPredictPositionConfig,
                         VizDoomTakeCoverConfig, VizDoomDeathMatchConfig, VizDoomHealthGatheringSupremeConfig]

    if selected_config is None:
        selected_config = np.random.choice(available_configs)

    agent = DeepQAgent.example(config=selected_config(agent_type='dqn', mode='stack'), n_episodes=20000,
                               max_episode_steps=10000, render=True, update_every=1, checkpoint_every=1000)
    agent.save()



from typing import Tuple

import gym
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from reinforcement_learning_keras.enviroments.preprocessing.clipper import Clipper


class ClipperWrapper(gym.ObservationWrapper):
    """Clip all observations to within limits, then StandardScale."""

    def __init__(self, env: gym.Env, lim: Tuple[float, float] = (-1, 1)):
        super().__init__(env)
        # New env obs space shape
        self.observation_space = gym.spaces.Box(low=lim[0], high=lim[1], shape=self.observation_space.shape)

        pipe = Pipeline([('clip', Clipper(lim=lim)),
                         ('ss', StandardScaler())])
        pipe.fit(np.array([[lim[0]] * self.observation_space.shape[0],
                           [lim[1]] * self.observation_space.shape[0]]))
        self.pp = pipe

    def observation(self, obs: np.ndarray) -> np.ndarray:
        return self.pp.transform(obs.reshape(1, -1))



import unittest

from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy


class TestEpsilonGreedy(unittest.TestCase):
    _sut = EpsilonGreedy
    _mock_attribute = 4

    def setUp(self) -> None:
        self._greedy_action = lambda: self._mock_attribute + 1
        self._random_action = lambda: 'random'

    def test_eps_1_returns_random_action_during_training(self):
        # Arrange
        eps = self._sut(eps_initial=1)

        # Act
        selection = eps.select(greedy_option=self._greedy_action,
                               random_option=self._random_action,
                               training=True)

        # Assert
        self.assertEqual('random', selection)

    def test_eps_1_returns_greedy_action_if_not_training(self):
        # Arrange
        eps = self._sut(eps_initial=1)

        # Act
        selection = eps.select(greedy_option=self._greedy_action,
                               random_option=self._random_action)

        # Assert
        self.assertEqual(5, selection)

    def test_eps_0_returns_greedy_action(self):
        # Arrange
        eps = self._sut(eps_initial=0)

        # Act
        selection = eps.select(greedy_option=self._greedy_action,
                               random_option=self._random_action)

        # Assert
        self.assertEqual(5, selection)

    def test_training_selection_decays_eps(self):
        # Arrange
        initial_eps = 0.99
        eps = self._sut(eps_initial=initial_eps)

        # Act
        _ = eps.select(greedy_option=self._greedy_action,
                       random_option=self._random_action,
                       training=True)

        # Assert
        self.assertLess(eps.eps_current, initial_eps)

    def test_non_training_selection_doesnt_decay_eps(self):
        # Arrange
        initial_eps = 0.99
        eps = self._sut(eps_initial=initial_eps)

        # Act
        _ = eps.select(greedy_option=self._greedy_action,
                       random_option=self._random_action,
                       training=False)

        # Assert
        self.assertAlmostEqual(eps.eps_current, initial_eps)

    def test_training_call_with_0_decay_doesnt_decay_eps(self):
        # Arrange
        initial_eps = 0.99
        eps = self._sut(eps_initial=initial_eps,
                        decay=0)

        # Act
        _ = eps.select(greedy_option=self._greedy_action,
                       random_option=self._random_action,
                       training=True)

        # Assert
        self.assertAlmostEqual(eps.eps_current, initial_eps)



"""Train and few DeepQAgents, plot the results, and run an episode on the best agent."""
from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from reinforcement_learning_keras.experiment.agent_experiment import AgentExperiment


def run_exp(agent_type: str, n_episodes: int = 1000, max_episode_steps: int = 600):
    gpu = VirtualGPU(256)

    exp = AgentExperiment(name=f"{agent_type} CartPole",
                          agent_class=DeepQAgent,
                          agent_config=CartPoleConfig(agent_type=agent_type),
                          n_reps=6,
                          n_jobs=1 if gpu.on else 3,
                          training_options={"n_episodes": n_episodes,
                                            "max_episode_steps": max_episode_steps})

    exp.run()
    exp.save(fn=f"{DeepQAgent.__name__}_{agent_type}experiment.pkl")


if __name__ == "__main__":
    run_exp(agent_type='dqn')
    run_exp(agent_type='dueling_dqn')
    run_exp(agent_type='double_dqn')
    run_exp(agent_type='double_dueling_dqn')



import copy
import warnings
from dataclasses import dataclass
from typing import Callable, Union, Dict, Any
from typing import List

import gym
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from joblib import Parallel, delayed
from joblib.externals.loky.process_executor import BrokenProcessPool

from reinforcement_learning_keras.agents.agent_base import AgentBase
from reinforcement_learning_keras.enviroments.config_base import ConfigBase


@dataclass
class AgentExperiment:
    agent_class: Callable
    agent_config: ConfigBase
    name: str = "unnamed_experiment"
    n_reps: int = 5
    n_jobs: int = 1
    training_options: Union[None, Dict[str, Any]] = None

    def __post_init__(self):
        self._trained_agents: List[AgentBase] = []
        self._set_default_training_options()

    def _set_default_training_options(self):
        if self.training_options is None:
            self.training_options = {}

        defaults = {"n_episodes": 500, "max_episode_steps": 500, "render": False, "verbose": False}
        for k, v in defaults.items():
            if k not in self.training_options:
                self.training_options[k] = defaults[k]

    @property
    def agent_scores(self) -> List[float]:
        return [a.training_history.current_performance for a in self._trained_agents]

    @property
    def best_agent(self) -> AgentBase:
        return self._trained_agents[int(np.argmax(self.agent_scores))]

    @property
    def worst_agent(self) -> AgentBase:
        return self._trained_agents[int(np.argmin(self.agent_scores))]

    @staticmethod
    def _fit_agent(agent_class: Callable, agent_config: ConfigBase, training_options: Dict[str, Any]):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FutureWarning)

            config_dict = agent_config.build()
            # Give each agent a unique name for easier tracking with verbose and multiprocessing
            config_dict["name"] = f"{config_dict.get('name', 'Agent')}_{np.random.randint(0, 2 ** 16)}"

            agent = agent_class(**config_dict)
            agent.train(**training_options)
            agent.unready()

        return agent

    def _run(self) -> None:
        self._trained_agents = Parallel(
            backend='loky', verbose=10,
            n_jobs=self.n_jobs)(delayed(self._fit_agent)(self.agent_class, self.agent_config, self.training_options)
                                for _ in range(self.n_reps))

    def run(self) -> None:
        try:
            # All agents support pickling, so joblib can run some in parallel...
            self._run()
        except BrokenProcessPool:
            # ... Except for TF models running on GPU, they'll probably crap out. Run 1 by 1.
            # OR it'll crash Chrome, and Python, and hang for all eternity. It's best not to rely on this.
            self.n_jobs = 1
            self._run()

        self.plot()
        self.play_best()

    def plot(self, err: str = 'range') -> None:
        """
        Plot reward vs episode for experiment.

        Plots:
        - Mean of all agents
        - Std or range of all agents (not constrained by max episodes or minimum score)
        - Min and max score across all agents for each episode (observed; constrained by max episodes or minimum score)
        - Score of best and worst agents (with 5% moving average). Best and worst defined using "current_performance"
          property of agents, which is mean score over most recent n episodes, where n is whatever the rolling average
          specified in the agents training history was.


        """

        sns.set()

        full_history = np.hstack([np.vstack(a.training_history.get_metric('total_reward'))
                                  for a in self._trained_agents])

        # Summary stats
        n_episodes = full_history.shape[0]
        y_mean = np.mean(full_history, axis=1)

        plt.plot(y_mean, color='#1f77b4', label='Mean score', lw=1.25)
        if err == 'range':
            plt.fill_between(range(n_episodes), np.min(full_history, axis=1), np.max(full_history, axis=1),
                             color='lightgrey', label='Score range', alpha=0.5)
        else:
            y_std = np.std(full_history, axis=1)
            plt.fill_between(range(n_episodes), y_mean - y_std, y_mean + y_std,
                             color='#1f77b4', label='Score std', alpha=0.3)

        # Best and worst agents
        mv_avg_pts = max(1, int(n_episodes * 0.05))  # 5% moving avg
        plt.plot(np.convolve(self.best_agent.training_history.get_metric('total_reward'),
                             np.ones(mv_avg_pts), 'valid') / mv_avg_pts,
                 label='Best (mv avg)', ls='--', color='#d62728', lw=0.7)
        plt.plot(np.convolve(self.worst_agent.training_history.get_metric('total_reward'),
                             np.ones(mv_avg_pts), 'valid') / mv_avg_pts,
                 label='Worst (mv avg)', ls='--', color='#9467bd', lw=0.7)

        plt.title(f'{self.name}', fontweight='bold')
        plt.xlabel('Episode', fontweight='bold')
        plt.ylabel('Score', fontweight='bold')
        plt.legend(title='Agents')
        plt.tight_layout()
        plt.savefig(f'{self.name}_{self.agent_config.env_spec}.png')
        plt.close()

    def play_best(self, episode_steps: int = None):
        if episode_steps is None:
            episode_steps = self.training_options["max_episode_steps"]

        best_agent = copy.deepcopy(self.best_agent)
        best_agent.check_ready()
        best_agent.env_builder.set_env(gym.wrappers.Monitor(best_agent.env,
                                                            f'{self._trained_agents[0].name}_monitor_dir',
                                                            force=True))

        try:
            best_agent.play_episode(training=False, render=False, max_episode_steps=episode_steps)
        except (ImportError, gym.error.DependencyNotInstalled) as e:
            print(f"Monitor wrapper failed, not saving video: \n{e}")

    def save(self, fn: str):
        """Disabled for now... Needed?"""
        pass

    def save_best_agent(self):
        self.best_agent.save()



import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

from reinforcement_learning_keras.enviroments.atari.environment_processing.frame_buffer_wrapper import \
    FrameBufferWrapper
from tests.unit.environments.pong.environment_processing.env_fixture import EnvFixture


class TestFrameBufferWrapper(unittest.TestCase):
    _sut = FrameBufferWrapper
    _env_fixture = EnvFixture

    def setUp(self):
        self._env = self._env_fixture()

    def test_reset_returns_expected_obs_values_and_shapes_with_stack_op(self):
        # Arrange
        env = self._sut(self._env, obs_shape=self._env.obs_shape)

        # Act
        obs = env.reset()

        # Assert
        self.assertEqual((self._env.obs_shape[0], self._env.obs_shape[1], 3),
                         obs.shape)
        self.assertEqual(np.unique(obs[:, :, 0:2]), 0)

    def test_step_returns_expected_obs_values_and_shapes_with_stack_op(self):
        # Arrange
        env = self._sut(self._env, obs_shape=self._env.obs_shape)

        # Act
        obs1 = env.reset()
        obs2, reward, done, _ = env.step(0)

        # Assert
        self.assertEqual((self._env.obs_shape[0], self._env.obs_shape[1], 3),
                         obs2.shape)
        self.assertEqual(np.unique(obs2[:, :, 0]), 0)
        assert_array_almost_equal(obs1[:, :, 2], obs2[:, :, 1])
        self.assertIsInstance(reward, float)
        self.assertIsInstance(done, bool)

    def test_reset_returns_expected_obs_values_and_shapes_with_diff_op(self):
        # Arrange
        env = self._sut(self._env, obs_shape=self._env.obs_shape,
                        buffer_length=2, buffer_function='diff')

        # Act
        obs = env.reset()

        # Assert
        self.assertEqual((self._env.obs_shape[0], self._env.obs_shape[1], 1),
                         obs.shape)
        self.assertEqual(np.unique(obs[:, :]), 1)

    def test_step_returns_expected_obs_values_and_shapes_with_diff_op(self):
        # Arrange
        env = self._sut(self._env, obs_shape=self._env.obs_shape,
                        buffer_length=2, buffer_function='diff')

        # Act
        _ = env.reset()
        obs2, reward, done, _ = env.step(0)
        # Assert
        self.assertEqual((self._env.obs_shape[0], self._env.obs_shape[1], 1),
                         obs2.shape)
        self.assertEqual(0, np.unique(obs2[:, :]))
        self.assertIsInstance(reward, float)
        self.assertIsInstance(done, bool)



import os
from functools import partial
from typing import Any, Dict, Tuple

# Although unused, this import will register Doom envs with Gym
# noinspection PyUnresolvedReferences
import vizdoomgym

from reinforcement_learning_keras.agents.components.replay_buffers.continuous_buffer import ContinuousBuffer
from reinforcement_learning_keras.agents.models.conv_nn import ConvNN
from reinforcement_learning_keras.agents.models.dueling_conv_nn import DuelingConvNN
from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy
from reinforcement_learning_keras.enviroments.config_base import ConfigBase
from reinforcement_learning_keras.enviroments.doom.environment_processing.frame_buffer_wrapper import FrameBufferWrapper
from reinforcement_learning_keras.enviroments.doom.environment_processing.image_process_wrapper import \
    ImageProcessWrapper


class DoomDefaultConfig(ConfigBase):
    """Defines default configs for Doom."""
    env_spec = 'VizdoomBasic-v0'
    n_actions = 3
    supported_agents = ('dqn', 'dueling_dqn', 'double_dqn', 'double_dueling_dqn', 'random')
    supported_modes = ('diff', 'stack')
    gpu_memory: int = 2048

    # 3 different possible resiltuons:
    # (96, 128) @ 40%
    # (96, 128) @ 20%
    # (90, 160) @ 20%
    res_scale: float = 0.4
    target_obs_shape: Tuple[int, int] = (96, 128)

    def __init__(self, mode: str = 'diff', *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if mode not in self.supported_modes:
            raise ValueError(f"Mode {mode} is not a supported mode ({self.supported_modes})")

        self._wrappers_stack = (partial(ImageProcessWrapper, scale=self.res_scale),
                                partial(FrameBufferWrapper, obs_shape=self.target_obs_shape, buffer_function='stack'))

        self._wrappers_diff = (partial(ImageProcessWrapper, scale=self.res_scale),
                               partial(FrameBufferWrapper, obs_shape=self.target_obs_shape, buffer_length=2,
                                       buffer_function='diff'))

        self.mode = mode
        if self.mode == "diff":
            self.env_wrappers = self._wrappers_diff
            self.frame_depth = 1

        if self.mode == "stack":
            self.env_wrappers = self._wrappers_stack
            self.frame_depth = 3

        self.wrapped_env = self.env_wrappers[1](self.env_wrappers[0](self.unwrapped_env))

    def _build_for_dqn(self) -> Dict[str, Any]:
        return {'name': os.path.join(self.folder, 'DeepQAgent'),
                'env_spec': self.env_spec,
                'env_wrappers': self.env_wrappers,
                'model_architecture': ConvNN(
                    observation_shape=(self.target_obs_shape[0], self.target_obs_shape[1], self.frame_depth),
                    n_actions=self.n_actions, output_activation=None, opt='adam', learning_rate=0.0001),
                'gamma': 0.99,
                'final_reward': None,
                'eps': EpsilonGreedy(eps_initial=1.2, decay=0.0000019, eps_min=0.01, decay_schedule='linear'),
                'replay_buffer': ContinuousBuffer(buffer_size=20000),
                'replay_buffer_samples': 32}

    def _build_for_dueling_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DuelingDQN'),
                            'model_architecture': DuelingConvNN(
                                observation_shape=(self.target_obs_shape[0], self.target_obs_shape[1],
                                                   self.frame_depth),
                                n_actions=6, opt='adam', learning_rate=0.0001)})

        return config_dict

    def _build_for_double_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DoubleDQN'),
                            'double': True,
                            'model_architecture': ConvNN(
                                observation_shape=(self.target_obs_shape[0], self.target_obs_shape[1],
                                                   self.frame_depth),
                                n_actions=self.n_actions, output_activation=None, opt='adam', learning_rate=0.0001)})

        return config_dict

    def _build_for_double_dueling_dqn(self) -> Dict[str, Any]:
        config_dict = self._build_for_dqn()
        config_dict.update({'name': os.path.join(self.folder, 'DoubleDuelingDQN'),
                            'double': True,
                            'model_architecture': DuelingConvNN(
                                observation_shape=(self.target_obs_shape[0], self.target_obs_shape[1],
                                                   self.frame_depth),
                                n_actions=self.n_actions, opt='adam', learning_rate=0.0001)})

        return config_dict

    def _build_for_random(self):
        return {'name': os.path.join(self.folder, 'RandomAgent'),
                'env_spec': self.env_spec}



import time
import unittest

import matplotlib.pyplot as plt
import numpy as np

from reinforcement_learning_keras.enviroments.atari.pong.pong_config import PongConfig


class TestPongStackEnvironment(unittest.TestCase):
    _sut = PongConfig(mode='stack', agent_type='dqn').wrapped_env
    _expected_shape = (84, 84, 3)
    _n_steps = 20
    _show = False

    def _plot_obs(self, obs: np.ndarray):
        n_buff = obs.shape[2]
        fig, ax = plt.subplots(ncols=n_buff)
        for i in range(n_buff):
            ax[i].imshow(obs[:, :, i])

        if self._show:
            fig.show()
            time.sleep(0.1)

    def test_reset_returns_expected_obs_shape(self):
        # Act
        obs = self._sut.reset()

        # Assert
        self.assertEqual(self._expected_shape, obs.shape)

    def test_reset_returns_expected_obs_value(self):
        # Act
        obs = self._sut.reset()

        # Assert
        self.assertLess(obs[0, 0, 0], 1)

    def test_step_returns_expected_obs_shape(self):
        # Arrange
        _ = self._sut.reset()

        # Act
        obs, reward, done, _ = self._sut.step(0)

        # Assert
        self.assertEqual(self._expected_shape, obs.shape)

    def test_step_returns_expected_obs_value(self):
        # Arrange
        _ = self._sut.reset()

        # Act
        obs, reward, done, _ = self._sut.step(0)

        # Assert
        self.assertLess(obs[0, 0, 0], 1)

    def test_multiple_steps(self):
        # Arrange
        _ = self._sut.reset()

        # Act
        for _ in range(self._n_steps):
            obs, reward, done, _ = self._sut.step(np.random.choice(range(self._sut.action_space.n)))

            # Manually assert
            self._plot_obs(obs)


class TestPongDiffEnvironment(TestPongStackEnvironment):
    _sut = PongConfig(mode='diff', agent_type='dqn').wrapped_env
    _expected_shape = (84, 84, 1)
    _show = False

    def _plot_obs(self, obs: np.ndarray):
        plt.imshow(obs.squeeze())

        if self._show:
            plt.show()
            time.sleep(0.1)



import os
import tempfile
import unittest

from reinforcement_learning_keras.agents.agent_base import AgentBase
from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.policy_gradient.reinforce_agent import ReinforceAgent
from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.agents.q_learning.linear_q_agent import LinearQAgent
from reinforcement_learning_keras.agents.random.random_agent import RandomAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from reinforcement_learning_keras.experiment.agent_experiment import AgentExperiment


class TestAgentExperiment(unittest.TestCase):
    _sut = AgentExperiment
    _agent_config = CartPoleConfig

    def setUp(self) -> None:
        self._tmp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self._tmp_dir.cleanup()

    def _run_exp(self, agent_class: AgentBase, agent_type: str, n_jobs: int = 1):
        # Arrange
        exp = AgentExperiment(name=os.path.join(self._tmp_dir.name, 'test_exp'),
                              agent_class=agent_class,
                              agent_config=self._agent_config(agent_type=agent_type, folder=self._tmp_dir.name),
                              n_reps=3,
                              n_jobs=n_jobs,
                              training_options={"n_episodes": 4,
                                                "max_episode_steps": 4})

        # Act
        exp.run()

        # Assert
        self.assertEqual(3, len(exp._trained_agents))
        self.assertEqual(3, len(exp.agent_scores))
        for a in exp._trained_agents:
            self.assertEqual(4, len(a.training_history.history))

    def test_short_random_agent_run_completes_with_expected_outputs(self):
        self._run_exp(RandomAgent, agent_type='random')

    def test_short_random_agent_parallel_run_completes_with_expected_outputs(self):
        self._run_exp(RandomAgent, agent_type='random', n_jobs=3)

    def test_short_linear_q_agent_run_completes_with_expected_outputs(self):
        self._run_exp(LinearQAgent, agent_type='linear_q')

    def test_short_linear_q_agent_parallel_run_completes_with_expected_outputs(self):
        self._run_exp(LinearQAgent, agent_type='linear_q', n_jobs=3)

    def test_short_deep_q_agent_run_completes_with_expected_outputs(self):
        VirtualGPU(256)
        self._run_exp(DeepQAgent, agent_type='dqn')

    def test_short_deep_q_agent_parallel_run_completes_with_expected_outputs(self):
        gpu = VirtualGPU(256)
        if not gpu:
            self._run_exp(DeepQAgent, agent_type='dqn')

    def test_short_reinforce_agent_run_completes_with_expected_outputs(self):
        VirtualGPU(256)
        self._run_exp(ReinforceAgent, agent_type='reinforce')

    def test_short_reinforce_agent_parallel_run_completes_with_expected_outputs(self):
        gpu = VirtualGPU(256)
        if not gpu:
            self._run_exp(ReinforceAgent, agent_type='reinforce')



import joblib


class ReplayBufferBase:

    def save(self, fn) -> None:
        joblib.dump(self, fn, compress=3)

    @classmethod
    def load(cls, fn: str) -> "ReplayBufferBase":
        return joblib.load(fn)



from typing import Tuple

from sklearn.base import BaseEstimator, TransformerMixin


class Clipper(BaseEstimator, TransformerMixin):
    lim: Tuple[float, float]

    def __init__(self, lim: Tuple[float, float] = (-1, 1)):
        self.set_params(lim=lim)

    def fit(self, x=None, y=None):
        return self

    def transform(self, x):
        x[x < self.lim[0]] = self.lim[0]
        x[x > self.lim[1]] = self.lim[1]
        return x



import unittest
from unittest.mock import call, Mock

import numpy as np

from reinforcement_learning_keras.enviroments.atari.environment_processing.max_and_skip_wrapper import MaxAndSkipWrapper
from tests.unit.environments.pong.environment_processing.env_fixture import EnvFixture


class TestMaxAndSkipWrapper(unittest.TestCase):
    _sut = MaxAndSkipWrapper
    _env_fixture = EnvFixture

    def setUp(self):
        self._env = self._env_fixture()

    def test_max_aggregation_of_state_returns_expected_step_outputs(self):
        # Arrange
        frame_buffer_length = 3
        n_action_frames = 6
        env = self._sut(self._env,
                        frame_buffer_length=frame_buffer_length, n_action_frames=n_action_frames)
        env.env.action_indicator = Mock()

        # Act
        _ = env.reset()
        obs, reward, done, _ = env.step(action=0)

        # Assert
        self.assertEqual(1, np.unique(obs))
        self.assertAlmostEqual(n_action_frames, reward)
        self.assertEqual(False, done)
        env.env.action_indicator.assert_has_calls(calls=[call(0) for _ in range(n_action_frames)])
        self.assertEqual(frame_buffer_length, len(env._frame_buffer))

    def test_sum_aggregation_of_state_returns_expected_step_outputs(self):
        # Arrange
        frame_buffer_length = 6
        n_action_frames = 8
        env = self._sut(self._env, frame_buffer_agg_f=np.sum,
                        frame_buffer_length=frame_buffer_length, n_action_frames=n_action_frames)
        env.env.action_indicator = Mock()

        # Act
        _ = env.reset()
        obs, reward, done, _ = env.step(action=-1)

        # Assert
        self.assertEqual(frame_buffer_length, np.unique(obs))
        self.assertAlmostEqual(n_action_frames, reward)
        self.assertEqual(False, done)
        env.env.action_indicator.assert_has_calls(calls=[call(-1) for _ in range(n_action_frames)])
        self.assertEqual(frame_buffer_length, len(env._frame_buffer))

    def test_aggregation_of_state_returns_expected_step_outputs_with_done_in_sequence(self):
        # Arrange
        frame_buffer_length = 5
        n_action_frames = 6
        env = self._sut(self._env, frame_buffer_length=frame_buffer_length,
                        n_action_frames=n_action_frames)
        env.env.action_indicator = Mock()

        # Act
        # Action 1 will cause fixture to return done after 2 turns. Done turn has additional reward.
        _ = env.reset()
        obs, reward, done, _ = env.step(action=1)

        # Assert
        self.assertEqual(2, np.unique(obs))
        self.assertAlmostEqual(4.0, reward)  # 3 steps +1 bonus reward
        self.assertEqual(True, done)
        env.env.action_indicator.assert_has_calls(calls=[call(1) for _ in range(3)])
        self.assertEqual(4, len(env._frame_buffer))  # 3 + 1 from the rest call, not 5



import tempfile
import unittest

from reinforcement_learning_keras.agents.random.random_agent import RandomAgent
from reinforcement_learning_keras.enviroments.atari.pong.pong_config import PongConfig
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig
from reinforcement_learning_keras.enviroments.mountain_car.mountain_car_config import MountainCarConfig


class TestRandomAgent(unittest.TestCase):
    _sut = RandomAgent
    _agent_type: str = 'random'

    def setUp(self) -> None:
        self._tmp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self._tmp_dir.cleanup()

    def test_saving_and_reloading_creates_identical_object(self):
        # Arrange
        agent = self._sut(**CartPoleConfig(agent_type=self._agent_type, plot_during_training=False,
                                           folder=self._tmp_dir.name).build())
        agent.train(verbose=True, render=False, n_episodes=2)

        # Act
        agent.save()
        agent_2 = self._sut.load(f"{agent.name}_{agent.env_spec}")
        agent_2.check_ready()

        # Assert
        self.assertEqual(agent, agent_2)

    def test_cart_pole_example(self):
        # Arrange
        config = CartPoleConfig(agent_type=self._agent_type, plot_during_training=False,
                                folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, n_episodes=200)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_mountain_car_example(self):
        # Arrange
        config = MountainCarConfig(agent_type=self._agent_type, plot_during_training=False,
                                   folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, max_episode_steps=100, n_episodes=20)

        # Assert
        self.assertIsInstance(agent, self._sut)

    def test_pong_example(self):
        # Arrange
        config = PongConfig(agent_type=self._agent_type, plot_during_training=False,
                            folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, max_episode_steps=100, n_episodes=20)

        # Assert
        self.assertIsInstance(agent, self._sut)



"""Train and few DeepQAgents, plot the results, and run an episode on the best agent."""
from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.policy_gradient.reinforce_agent import ReinforceAgent
from reinforcement_learning_keras.enviroments.mountain_car.mountain_car_config import MountainCarConfig
from reinforcement_learning_keras.experiment.agent_experiment import AgentExperiment


def run_exp(n_episodes: int = 1000, max_episode_steps: int = 1000):
    gpu = VirtualGPU(256)

    exp = AgentExperiment(agent_class=ReinforceAgent,
                          agent_config=MountainCarConfig(agent_type='reinforce'),
                          n_reps=5,
                          n_jobs=1 if gpu.on else 5,
                          training_options={"n_episodes": n_episodes,
                                            "max_episode_steps": max_episode_steps,
                                            "update_every": 1})

    exp.run()
    exp.save(fn=f"{ReinforceAgent.__name__}_experiment.pkl")


if __name__ == "__main__":
    run_exp(n_episodes=1000)



import gc
import tempfile
import unittest

import tensorflow as tf

from reinforcement_learning_keras.agents.components.helpers.virtual_gpu import VirtualGPU
from reinforcement_learning_keras.agents.policy_gradient.reinforce_agent import ReinforceAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig


class TestReinforceAgent(unittest.TestCase):
    _sut = ReinforceAgent
    _agent_type: str = 'reinforce'
    _gpu = VirtualGPU(256)

    def setUp(self) -> None:
        self._tmp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        tf.keras.backend.clear_session()
        tf.compat.v1.reset_default_graph()
        gc.collect()
        self._tmp_dir.cleanup()

    def test_saving_and_reloading_creates_identical_object(self):
        # Arrange
        agent = self._sut(**CartPoleConfig(agent_type=self._agent_type, plot_during_training=False,
                                           folder=self._tmp_dir.name).build())
        agent.train(verbose=True, render=False, n_episodes=2)

        # Act
        agent.save()
        agent_2 = self._sut.load(f"{agent.name}_{agent.env_spec}")
        agent_2.check_ready()

        # Assert
        self.assertEqual(agent, agent_2)

    def test_cart_pole_example(self):
        # Arrange
        config = CartPoleConfig(agent_type=self._agent_type, plot_during_training=False,
                                folder=self._tmp_dir.name)

        # Act
        agent = self._sut.example(config, render=False, n_episodes=10)

        # Assert
        self.assertIsInstance(agent, self._sut)



from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.enviroments.atari.space_invaders.space_invaders_config import SpaceInvadersConfig

if __name__ == "__main__":
    agent = DeepQAgent.example(config=SpaceInvadersConfig(agent_type='dueling_dqn', mode='stack'),
                               max_episode_steps=4000, n_episodes=10000,
                               render=False, update_every=2, checkpoint_every=2000)
    agent.save()



from dataclasses import dataclass
from functools import reduce
from typing import Union, Callable, Iterable

import gym


@dataclass
class EnvBuilder:
    env_spec: str
    env_wrappers: Iterable[Callable] = None

    def __post_init__(self):
        self._env: Union[None, gym.Env] = None
        if self.env_wrappers is None:
            self.env_wrappers = []

        self._register_other_envs()
        self.set_env()

    @staticmethod
    def _register_other_envs() -> None:
        """Try and import supported envs (may not be installed). Need to do this to register them with gym."""
        try:
            import vizdoomgym
        except ImportError:
            pass

    def set_env(self, env: Union[None, gym.Env] = None) -> None:
        """
        Create a new env object from the spec, or set a new one.

        Can specify a new env, this is useful, for example, to add a Monitor wrapper.
        """

        if env is not None:
            self._env = env

        if self._env is None:
            # Make the gym environment and apply the wrappers one by one
            self._env = reduce(lambda inner_env, wrapper: wrapper(inner_env),
                               self.env_wrappers,
                               gym.make(self.env_spec))

    @property
    def env(self) -> gym.Env:
        """Use to access env, if not ready also makes it ready."""
        self.set_env()

        return self._env



from dataclasses import dataclass

import numpy as np

from reinforcement_learning_keras.agents.q_learning.exploration.epsilon_greedy import EpsilonGreedy


@dataclass
class EpisodeReport:
    frames: int
    time_taken: float
    total_reward: float
    epsilon_used: EpsilonGreedy = None

    def __str__(self) -> str:
        return f"Reward: {self.total_reward} from {self.frames} frames in {self.time_taken} s ({self.fps} f/s). " \
               f"Eps remaining: {self.epsilon_used.eps_current if self.epsilon_used is not None else None}"

    @property
    def fps(self) -> float:
        return np.round(self.frames / self.time_taken)



import collections
from typing import Any, Dict, Tuple

import gym
import numpy as np


class FrameBufferWrapper(gym.Wrapper):
    """
    Adds last step obs to buffer, returns whole buffer.

    Returned buffer contains previous steps, eg. for env.step at t=3, returns obs for t=3, t=2, t=1. In games like pong,
    this adds directional information to the returned observations.
    """

    def __init__(self, env: gym.Env,
                 obs_shape: Tuple[int, int] = (84, 84),
                 buffer_length: int = 3,
                 buffer_function: str = 'stack') -> None:
        """
        :param env: Gym env.
        :param obs_shape: Expected shape of single observation.
        :param buffer_length: Number of frames to include in buffer.
        :param buffer_function: Function to apply to use contents of buffer. Supports 'stack or 'diff':
                                  - 'stack' stack contents of buffer on a new (final) axis
                                  - 'diff' take diff between two frames without changing dimensions.
        """
        super().__init__(env)
        self._buffer_length = buffer_length
        self._buffer_function = buffer_function
        self._obs_shape = obs_shape
        self._prepare_obs_buffer()

    def _prepare_obs_buffer(self) -> None:
        """Create buffer and preallocate with empty arrays of expected shape."""

        self._obs_buffer = collections.deque(maxlen=self._buffer_length)

        for _ in range(self._buffer_length):
            self._obs_buffer.append(np.zeros(shape=self._obs_shape))

    def _buffer_obs(self) -> np.ndarray:
        agg_buff = None

        if self._buffer_function == "stack":
            agg_buff = np.stack([obs.squeeze() for obs in self._obs_buffer],
                                axis=len(self._obs_shape))

        if self._buffer_function == 'diff':
            if self._buffer_length != 2:
                raise ValueError("When using diff, buffer length must be 2.")
            agg_buff = self._obs_buffer[1].squeeze() - self._obs_buffer[0].squeeze()
            agg_buff = np.expand_dims(agg_buff,
                                      axis=len(self._obs_shape))

        if agg_buff is None:
            raise ValueError(f"Unknown buffer op {self._buffer_function}")

        return agg_buff

    def step(self, action: int) -> Tuple[np.ndarray, float, bool, Dict[Any, Any]]:
        """Step env, add new obs to buffer, return buffer."""
        obs, reward, done, info = self.env.step(action)
        self._obs_buffer.append(obs)

        return self._buffer_obs(), reward, done, info

    def reset(self) -> np.ndarray:
        """Add initial obs to end of pre-allocated buffer.

        :return: Buffered observation
        """
        self._prepare_obs_buffer()
        obs = self.env.reset()
        self._obs_buffer.append(obs)

        return self._buffer_obs()



import abc
import copy
import gc
import time
from typing import Any, Callable, Union, Dict, Tuple, Iterable

import gym
import joblib
import numpy as np

from reinforcement_learning_keras.agents.components.helpers.env_builder import EnvBuilder
from reinforcement_learning_keras.agents.components.helpers.tqdm_handler import TQDMHandler
from reinforcement_learning_keras.agents.components.history.episode_report import EpisodeReport
from reinforcement_learning_keras.agents.components.history.training_history import TrainingHistory


class AgentBase(abc.ABC):
    name: str
    env_spec: str
    env_builder: Union[EnvBuilder, None]
    env_wrappers: Iterable[Callable]
    gamma: float
    final_reward: float
    ready: bool

    training_history: TrainingHistory
    _tqdm = TQDMHandler()

    @property
    def env(self) -> gym.Env:
        return self.env_builder.env

    def _pickle_compatible_getstate(self) -> Dict[str, Any]:
        """
        Prepare agent with a keras model object for pickling.

        Calls .unready to prepare this object for pickling, and .check_ready to put it back how it was after pickling.
        In addition to what's defined in the unready method, it avoids trying to copy the env. We can't copy this,
        but we also don't want to make a new one (like we can with the models). This would cause a new render window
        per episode...

        The default unready method in AgentBase does nothing. The GPU models should modify .unready and .check_ready to
        handle complied Keras, which also can't be pickled.

        Object that need to use this should implement their own __getstate__:
        def __getstate__(self) -> Dict[str, Any]:
            return self._pickle_compatible_getstate()
        It's not implemented in AgentBase as the standard __getstate__ is required by the deepcopy below.
        """

        # Remove things
        self.unready()

        # Get object spec to pickle, everything left except env_builder references
        # This is dodgy. Can end up with recursion errors depending on how deepcopy behaves...
        object_state_dict = copy.deepcopy({k: v for k, v in self.__dict__.items()})

        # Put this object back how it was
        self.check_ready()

        return object_state_dict

    def check_ready(self) -> None:
        """
        Check the model is ready to use.

        If super is used, should be at end of overloading method.

        Default implementation:
         - Check _env is set (most models?)
        Example of other model specific steps that might need doing:
         - For Keras models, check model is ready, for example if it needs recompiling after loading.
        """
        self.env_builder = EnvBuilder(self.env_spec, self.env_wrappers)
        self.env_builder.set_env()
        self.ready = True
        gc.collect()

    def unready(self) -> None:
        """
        Remove anything that causes issues with pickling, such as keras models.

        If super is used, should be at end of overloading method.
        """
        if self.env_builder is not None:
            self.env_builder.env.close()
            self.env_builder = None
        self.ready = False
        gc.collect()

    def transform(self, s: Any) -> Any:
        """Run the any pre-preprocessing on raw state, if used."""
        return s

    def update_experience(self, *args) -> None:
        """Remember an experience, if used by agent."""
        pass

    @abc.abstractmethod
    def _build_model(self) -> None:
        """Prepare the model(s) the agent will use."""
        pass

    @abc.abstractmethod
    def update_model(self, *args, **kwargs) -> None:
        """Update the agents model(s)."""
        pass

    def _discounted_reward(self, reward: float, estimated_future_action_rewards: np.ndarray) -> float:
        """Use this to define the discounted reward for unfinished episodes, default is 1 step TD."""
        return reward + self.gamma * np.max(estimated_future_action_rewards)

    def _get_reward(self, reward: float, estimated_future_action_rewards: np.ndarray, done: bool) -> float:
        """
        Calculate discounted reward for a single step.

        :param reward: Last real reward.
        :param estimated_future_action_rewards: Estimated future values of actions taken on next step.
        :param done: Flag indicating if this is the last step on an episode.
        :return: Reward.
        """

        if done:
            # If done, reward is just this step. Can finish because agent has won or lost.
            return self._final_reward(reward)
        else:
            # Otherwise, it's the reward plus the predicted max value of next action
            return self._discounted_reward(reward, estimated_future_action_rewards)

    @abc.abstractmethod
    def get_action(self, s: Any, **kwargs) -> int:
        """
        Given state s, get an action from the agent.

        May include other kwargs if needed - for example, a training flag for methods using epsilon greedy.
        """
        pass

    @abc.abstractmethod
    def _play_episode(self, max_episode_steps: int = 500,
                      training: bool = False, render: bool = True) -> Tuple[float, int]:
        """
        Play a single episode with the agent (run multiple steps). Should return reward and n frames.

        :param max_episode_steps: Max steps before stopping, overrides any time limit set by Gym.
        :param training: Bool to indicate whether or not to use this experience to update the model.
        :param render: Bool to indicate whether or not to call env.render() each training step.
        :return: The total real reward for the episode and number of frames run.
        """
        pass

    def play_episode(self, max_episode_steps: int = 500,
                     training: bool = False, render: bool = True) -> EpisodeReport:
        """
        Run Agent's _play_episode and produce episode report.

        :param max_episode_steps: Max steps before stopping, overrides any time limit set by Gym.
        :param training: Bool to indicate whether or not to use this experience to update the model.
        :param render: Bool to indicate whether or not to call env.render() each training step.
        :return: The total real reward for the episode, number of frames run, and time taken in seconds.
        """
        t0 = time.time()
        total_reward, frames = self._play_episode(max_episode_steps=max_episode_steps, training=training, render=render)
        t1 = time.time()

        return EpisodeReport(total_reward=total_reward,
                             frames=frames,
                             time_taken=np.round(t1 - t0, 3),
                             epsilon_used=getattr(self, 'eps', None))

    def train(self, n_episodes: int = 10000, max_episode_steps: int = 500, verbose: bool = True, render: bool = True,
              checkpoint_every: Union[bool, int] = 0, update_every: Union[bool, int] = 1) -> None:
        """
        Run the default training loop

        :param n_episodes: Number of episodes to run.
        :param max_episode_steps: Max steps before stopping, overrides any time limit set by Gym.
        :param verbose:  If verbose, use tqdm and print last episode score for feedback during training.
        :param render: Bool to indicate whether or not to call env.render() each training step.
        :param checkpoint_every: Save the model every n steps while training. Set to 0 or false to turn off.
        :param update_every: Run the _after_episode_update() step every n episodes.
        """
        self._tqdm.set_tqdm(verbose)

        for ep in self._tqdm.tqdm_runner(range(n_episodes)):
            episode_report = self.play_episode(max_episode_steps=max_episode_steps, training=True, render=render)
            self._update_history(episode_report, verbose)

            if (update_every > 0) and not (ep % update_every):
                # Run the after-episode update step
                self._after_episode_update()

            if (checkpoint_every > 0) and not (ep % checkpoint_every):
                self.save()

    def _after_episode_update(self) -> None:
        """
        Run an update step after an episode completes.

        In the default implementation of .train, update_every parameter can be used to control how often this method
        runs.

        Eg.
         - For DQN, synchronize target and value models
         - For REINFORCE do MC model training step
         - For random agent this is passed

         Note this there is no equivalent "during episode update" method as each agent defines it's own play_episode
         method, which can call model specific updates as needed (eg.f or DQN this sample from buffer + training
         step, for REINFORCE do nothing, etc.)
        """
        pass

    def _update_history(self, episode_report: EpisodeReport, verbose: bool = True) -> None:
        """
        Add an episodes reward to history and maybe plot depending on history settings.

        :param episode_report: Episode report to add to history.
        :param verbose: If verbose, print the last episode and run the history plot. The history plot will display
                        depending on it's own settings. Verbose = False will turn it off totally.
        """
        self.training_history.append(episode_report)

        if verbose:
            print(f"{self.name}: {episode_report}")
            self.training_history.training_plot()

    def save(self) -> None:
        with open(f"{self.name}_{self.env_spec}", 'wb') as f:
            joblib.dump(self, f)

    @classmethod
    def load(cls, fn: str) -> "AgentBase":
        with open(fn, 'rb') as f:
            new_agent = joblib.load(f)
        new_agent.check_ready()

        return new_agent

    @classmethod
    def example(cls, config: Dict[str, Any]) -> "AgentBase":
        """Optional example function using this agent."""
        raise NotImplementedError



from dataclasses import dataclass
from typing import List, Tuple, Any

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from reinforcement_learning_keras.agents.components.history.episode_report import EpisodeReport


@dataclass
class TrainingHistory:
    plotting_on: bool = False
    plot_every: int = 50
    agent_name: str = 'Unnamed agents'
    rolling_average: int = 10

    def __post_init__(self) -> None:
        sns.set()
        self.history: List[EpisodeReport] = []

    def append(self, episode_report: EpisodeReport) -> None:
        self.history.append(episode_report)

    def extend(self, episode_report: List[EpisodeReport]) -> None:
        self.history.extend(episode_report)

    def get_metric(self, metric: str = "total_reward") -> List[Any]:
        return [getattr(ep, metric) for ep in self.history]

    def plot(self, metrics: List[str], show: bool = True) -> Tuple[plt.Figure, plt.Axes]:
        """Plot current history."""
        return self._plot(show=show, metrics=metrics)

    def training_plot(self, show: bool = True) -> Tuple[plt.Figure, plt.Axes]:
        """Plot if it's turned on and is a plot step."""

        if self.plotting_on and (not len(self.history) % self.plot_every):
            return self._plot(show=show, metrics=["total_reward", "frames"])

    def _plot(self, metrics: List[str], show: bool = True) -> Tuple[plt.Figure, plt.Axes]:
        """Create axes and plot. Not storing matplotlib objects to self as they cause pickle issues."""
        plt.close('all')
        fig, axs = plt.subplots(nrows=len(metrics), ncols=1)
        if not isinstance(axs, np.ndarray):
            axs = [axs]

        for ax, metric in zip(axs, metrics):
            ax.plot(self._rolling_average(self.get_metric(metric)), label=metric)
            ax.set_xlabel('N Episodes', fontweight='bold')
            ax.set_ylabel(metric, fontweight='bold')
        axs[0].set_title(self.agent_name, fontweight='bold')

        if show:
            fig.show()

        return fig, ax

    def _rolling_average(self, x) -> np.ndarray:
        """Rolling average over """
        return np.convolve(x, np.ones(self.rolling_average), 'valid') / self.rolling_average

    @property
    def total_frames(self) -> int:
        return int(np.sum(self.get_metric("frames")))

    @property
    def current_performance(self, metric: str = "total_reward") -> float:
        """Return average performance over the last rolling average window."""
        return float(np.mean(self.get_metric(metric)[len(self.history) - self.rolling_average:-1]))



from reinforcement_learning_keras.agents.q_learning.deep_q_agent import DeepQAgent
from reinforcement_learning_keras.enviroments.cart_pole.cart_pole_config import CartPoleConfig

if __name__ == "__main__":
    agent = DeepQAgent.example(config=CartPoleConfig(agent_type='double_dqn'), max_episode_steps=10000, render=False,
                               update_every=3, checkpoint_every=0)
    agent.save()



import setuptools

from reinforcement_learning_keras import __version__

setuptools.setup(
    name="reinforcement_learning_keras",
    version=__version__,
    author="Gareth Jones",
    author_email="garethgithub@gmail.com",
    description="",
    long_description='',
    long_description_content_type="text/markdown",
    url="https://github.com/garethjns/reinforcement-learning-keras",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"],
    python_requires='>=3.6',
    install_requires=["tensorflow==2.2.0", "scikit-learn==0.23.0", "matplotlib", "gym[atari]==0.17.1",
                      "dataclasses", "tqdm", "seaborn", "joblib", "numpy", "coverage", "mock", "opencv-python",
                      "joblib"])



import abc
from typing import List, Dict, Any

import gym

from reinforcement_learning_keras.agents.components.history.training_history import TrainingHistory


class ConfigBase(abc.ABC):
    env_spec: str
    supported_agents: List[str]
    gpu_memory: int = 256

    def __init__(self, agent_type: str, plot_during_training: bool = True, folder: str = ''):
        self.plot_during_training = plot_during_training
        self.folder = folder

        # Example env
        self.unwrapped_env = gym.make(self.env_spec)
        self.wrapped_env = self.unwrapped_env  # (Default no wrapper)

        self._check_supported(agent_type)
        self.agent_type = agent_type

    def _check_supported(self, agent_type: str):
        if agent_type not in self.supported_agents:
            raise NotImplementedError(f"Agent {agent_type} not in supported agents: {self.supported_agents}")

    @property
    def _default_training_history_kwargs(self) -> Dict[str, Any]:
        return {"plotting_on": self.plot_during_training, "plot_every": 50, "rolling_average": 25}

    def build(self) -> Dict[str, Any]:

        if self.agent_type.lower() == 'linear_q':
            config_dict = self._build_for_linear_q()
        elif self.agent_type.lower() == 'dqn':
            config_dict = self._build_for_dqn()
        elif self.agent_type.lower() == 'dueling_dqn':
            config_dict = self._build_for_dueling_dqn()
        elif self.agent_type.lower() == 'double_dqn':
            config_dict = self._build_for_double_dqn()
        elif self.agent_type.lower() == 'double_dueling_dqn':
            config_dict = self._build_for_double_dueling_dqn()
        elif self.agent_type.lower() == 'reinforce':
            config_dict = self._build_for_reinforce()
        elif self.agent_type.lower() == 'random':
            config_dict = self._build_for_random()
        else:
            raise NotImplementedError

        config_dict.update({'training_history': TrainingHistory(agent_name=config_dict['name'],
                                                                **self._default_training_history_kwargs)})
        return config_dict

    def _build_for_linear_q(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _build_for_dqn(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _build_for_dueling_dqn(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _build_for_double_dqn(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _build_for_double_dueling_dqn(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _build_for_random(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _build_for_reinforce(self) -> Dict[str, Any]:
        raise NotImplementedError



from typing import Any, Dict, Tuple

import gym
import numpy as np


class FireStartWrapper(gym.Wrapper):
    def __init__(self, env: gym.Env, fire_action_id: int = 1):
        super().__init__(env)
        self.fire_action_id = fire_action_id

    def step(self, action: int) -> Tuple[np.ndarray, float, bool, Dict[Any, Any]]:
        return self.env.step(action)

    def reset(self) -> np.ndarray:
        self.env.reset()
        obs, _, done, _ = self.env.step(self.fire_action_id)

        return obs



from typing import Tuple

from tensorflow import keras
from tensorflow.keras import backend as K

from reinforcement_learning_keras.enviroments.model_base import ModelBase


class DuelingConvNN(ModelBase):
    """A convolutional NN for Pong, similar to Google paper."""

    def _model_architecture(self) -> Tuple[keras.layers.Layer, keras.layers.Layer]:
        n_units = 512 * self.unit_scale

        frame_input = keras.layers.Input(name='input', shape=self.observation_shape)
        conv1 = keras.layers.Conv2D(32, kernel_size=(8, 8), strides=(4, 4),
                                    name='conv1', padding='same', activation='relu')(frame_input)
        conv2 = keras.layers.Conv2D(64, kernel_size=(4, 4), strides=(2, 2),
                                    name='conv2', padding='same', activation='relu')(conv1)
        conv3 = keras.layers.Conv2D(64, kernel_size=(3, 3), strides=(1, 1),
                                    name='conv3', padding='same', activation='relu')(conv2)
        flatten = keras.layers.Flatten(name='flatten')(conv3)
        fc1 = keras.layers.Dense(units=int(n_units), name='fc1', activation='relu')(flatten)
        # Separate layers for baseline value (1 node) and action advantages (n action nodes)
        v_layer = keras.layers.Dense(1, activation='linear')(fc1)
        a_layer = keras.layers.Dense(self.n_actions, activation='linear')(fc1)

        # Function definition for lambda layer. Base value layer + actions value layer - mean action values
        def _merge_layer(layer_inputs: list):
            return layer_inputs[0] + layer_inputs[1] - K.mean(layer_inputs[1], axis=1, keepdims=True)

        action_output = keras.layers.Lambda(_merge_layer, output_shape=(self.n_actions,),
                                            name="output")([v_layer, a_layer])

        return frame_input, action_output



import unittest

from reinforcement_learning_keras.enviroments.atari.environment_processing.image_process_wrapper import \
    ImageProcessWrapper
from tests.unit.environments.pong.environment_processing.env_fixture import EnvFixture


class TestImageProcessWrapper(unittest.TestCase):
    _sut = ImageProcessWrapper
    _env_fixture = EnvFixture

    def setUp(self):
        self._env = self._env_fixture(obs_shape=(250, 160, 3))

    def test_image_resize_from_250_160_3_to_expected_shape_on_reset(self):
        # Arrange
        env = self._sut(self._env)

        # Act
        obs = env.reset()

        # Assert
        self.assertEqual((84, 84), obs.shape)
        self.assertLess(obs[0, 0], 1)

    def test_image_resize_from_250_160_3_to_expected_shape_on_step(self):
        # Arrange
        env = self._sut(self._env)

        # Act
        obs, reward, done, _ = env.step(0)

        # Assert
        self.assertEqual((84, 84), obs.shape)
        self.assertLess(obs[0, 0], 1)

    def test_raises_error_with_incompatible_res(self):
        env = self._sut(self._env_fixture(obs_shape=(100, 100, 3)))

        # Act
        self.assertRaises(ValueError, lambda: env.step(0))
